# -*- coding: utf-8 -*-
"""
This is a template for a pysat.Instrument support file.
Modify this file as needed when adding a new Instrument to pysat.

This is a good area to introduce the instrument, provide background
on the mission, operations, instrumentation, and measurements.

Also a good place to provide contact information. This text will
be included in the pysat API documentation.

Properties
----------
platform
    *List platform string here*
name
    *List name string here*
inst_id
    *List supported inst_ids here*
tag
    *List supported tag strings here*

Note
----
- Optional section, remove if no notes

Warnings
--------
- Optional section, remove if no warnings
- Two blank lines needed afterward for proper formatting


Examples
--------
::

    Example code can go here


Authors
-------
Author name and institution

"""

import datetime as dt

import pysat

# Assign the pysat logger to the local log commands, as these functions will
# all be executed within pysat.  If this is the only instance pysat is used,
# consider omitting the pysat import and logger assignment and replacing it
# with:
# from pysat import logger
logger = pysat.logger

# ----------------------------------------------------------------------------
# Instrument attributes:

# The platform and name strings associated with this instrument need to be
# defined at the top level.  These attributes will be copied over to the
# Instrument object by pysat.  The strings used here should also be used to
# name this file platform_name.py
platform = ''
name = ''

# Dictionary of data 'tags' and corresponding description
tags = {'': 'description 1',  # this is the default
        'tag_string': 'description 2'}

# Let pysat know if there are multiple satellite platforms supported
# by these routines.  Define a dictionary keyed by instrument ID, each with a
# list of corresponding tags.  For example:
# inst_ids = {'a': ['tag1', 'tag2'], 'b': ['tag2', 'tag3']}
inst_ids = {'': ['', 'tag_string']}

# Set to False to specify using xarray (not using pandas)
# Set to True if data will be returned via a pandas DataFrame
pandas_format = False

# The following attributes will be set to these default values upon
# instantiation if not otherwise specified
directory_format = None
file_format = None
multi_file_day = False  # Set to True if files span less than a day
orbit_info = None

# ----------------------------------------------------------------------------
# Instrument testing attributes:

# Define good days to download data for when pysat undergoes testing.
# format is outer dictionary has inst_id as the key
# each inst_id has a dictionary of test dates keyed by tag string
# _test_dates = {'a':{'tag1': dt.datetime(2019,1,1),
#                     'tag2': dt.datetime(2019,1,1)},
#                'b':{'tag2': dt.datetime(2019,1,1),
#                     'tag3': dt.datetime(2019,1,1),}}
_test_dates = {'': {'': dt.datetime(2019, 1, 1),
                    'tag_string': dt.datetime(2019, 1, 2)}}

# Set Testing flags
# Dict structure should mirror _test_dates above

# For instruments without download support, set _test_download to False
# If not set, defaults to True
_test_download = {'': {'': False, 'tag_string': True}}

# For instruments using FTP for download, set _test_download_travis to False
# These tests will still download locally but be skipped on Travis CI
# If not set, defaults to True
_test_download_travis = {'': {'': False, 'tag_string': False}}

# For instruments requiring a user passwrod, set _password_req to True
# These instruments will not be downloaded as part of tests to preserve password
# security
# If not set, defaults to False
_password_req = {'': {'': True, 'tag_string': False}}

# ----------------------------------------------------------------------------
# Instrument methods: routines that are attached to the pysat.Instrument
# as class methods


# Required method
def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation. Object modified in place.  Use this to set
    the acknowledgements and references.

    """

    # Required attribute: acknowledgements
    self.acknowledgements = 'This would go in the Acknowledgements section'

    # Required attribute: references
    self.references = 'These are the instrument references'

    # Direct feedback to logging info
    logger.info(self.acknowledgements)
    return


# Required method
def clean(self):
    """Method to return PLATFORM/NAME data cleaned to the specified level

    Cleaning level is specified in inst.clean_level and pysat
    will accept user input for several strings. The clean_level is
    specified at instantiation of the Instrument object, though it may be
    updated to a more stringent level and re-applied after instantiation.
    The clean method is applied after default every time data is loaded.

    Note
    ----
    - 'clean' All parameters are good, suitable for scientific studies
    - 'dusty' Most parameters are good, requires instrument familiarity
    - 'dirty' There are data areas that have issues, use with caution
    - 'none' No cleaning applied, routine not called in this case.

    """

    return


# Optional method
def preprocess(self):
    """Customization method that performs standard preprocessing.

    This routine is automatically applied to the Instrument object
    on every load by the pysat nanokernel (first in queue). Object
    modified in place.

    """

    return

# ----------------------------------------------------------------------------
# Instrument functions: routines that are attached to the pysat.Instrument
# as function attributes


# Required function
def list_files(tag=None, inst_id=None, data_path=None, format_str=None):
    """Produce a list of files corresponding to PLATFORM/NAME.

    This routine is invoked by pysat and is not intended for direct
    use by the end user. Arguments are provided by pysat.

    Parameters
    ----------
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. (default='')
    inst_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. (default='')
    data_path : string
        Full path to directory containing files to be loaded. This
        is provided by pysat. The user may specify their own data path
        at Instrument instantiation and it will appear here. (default=None)
    format_str : string
        String template used to parse the datasets filenames. If a user
        supplies a template string at Instrument instantiation
        then it will appear here, otherwise defaults to None. (default=None)

    Returns
    -------
    pandas.Series
        Series of filename strings, including the path, indexed by datetime.

    Examples
    --------
    ::

        If a filename is SPORT_L2_IVM_2019-01-01_v01r0000.NC then the template
        is 'SPORT_L2_IVM_{year:04d}-{month:02d}-{day:02d}_' +
        'v{version:02d}r{revision:04d}.NC'


    Note
    ----
    The returned Series should not have any duplicate datetimes. If there are
    multiple versions of a file the most recent version should be kept and the
    rest discarded. This routine uses the pysat.Files.from_os constructor, thus
    the returned files are up to pysat specifications.

    Multiple data levels may be supported via the 'tag' input string.
    Multiple instruments via the inst_id string.

    """

    if format_str is None:
        # user did not supply an alternative format template string
        format_str = 'example_name_{year:04d}_{month:02d}_{day:02d}.nc'
    # we use a pysat provided function to grab list of files from the
    # local file system that match the format defined above
    file_list = pysat.Files.from_os(data_path=data_path, format_str=format_str)

    return file_list


# Required function
def download(date_array, tag, inst_id, data_path=None, user=None, password=None,
             **kwargs):
    """Placeholder for PLATFORM/NAME downloads.

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need not
        be contiguous.
    tag : string
        Tag identifier used for particular dataset. This input is provided by
        pysat. (default='')
    inst_id : string
        Satellite ID string identifier used for particular dataset. This input
        is provided by pysat. (default='')
    data_path : string
        Path to directory to download data to. (default=None)
    user : string (OPTIONAL)
        User string input used for download. Provided by user and passed via
        pysat. If an account is required for dowloads this routine here must
        error if user not supplied. (default=None)
    password : string (OPTIONAL)
        Password for data download. (default=None)
    custom_keywords : placeholder
        Additional keywords supplied by user when invoking the download
        routine attached to a pysat.Instrument object are passed to this
        routine. Use of custom keywords here is discouraged.

    """

    return


# Required function
def load(fnames, tag=None, inst_id=None, custom_keyword=None):
    """Loads PLATFORM data into (PANDAS/XARRAY).

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    fnames : array-like
        iterable of filename strings, full path, to data files to be loaded.
        This input is nominally provided by pysat itself.
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. While
        tag defaults to None here, pysat provides '' as the default
        tag unless specified by user at Instrument instantiation. (default='')
    inst_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. (default='')
    custom_keyword : type to be set
        Developers may include any custom keywords, with default values
        defined in the method signature. This is included here as a
        place holder and should be removed.

    Returns
    -------
    data, metadata
        Data and Metadata are formatted for pysat. Data is a
        pandas DataFrame or xarray DataSet while metadata is a pysat.Meta
        instance.

    Note
    ----
    Any additional keyword arguments passed to pysat.Instrument
    upon instantiation are passed along to this routine.

    Examples
    --------
    ::

        inst = pysat.Instrument('ucar', 'tiegcm')
        inst.load(2019, 1)

    """

    # netCDF4 files, particularly those produced by pysat can be loaded using a
    # pysat provided function, load_netcdf4.

    # Metadata across different files can employ different labels for the
    # same category of information. In our notional example, the file is
    # from NASA's CDAWeb and uses their standard for metadata. The dictionary
    # passed into the labels keyword is keyed by the category type that pysat
    # tracks, while the value contains a tuple for the string label used
    # within the file as well as the type of metadata to be tracked.
    # `load_netcdf4` uses this information to correctly parse metadata within
    # the file and attach it as appropriate. Once both the data and
    # metadata are returned by this function, pysat will adapt the metadata
    # labels to match the standard chosen by the user, not the
    # labels chosen by CDAWeb.
    data, mdata = pysat.utils.load_netcdf4(fnames, epoch_name='Epoch',
                                           labels={'units': ('Units', str),
                                                   'name': ('Long_Name', str),
                                                   'notes': ('Var_Notes', str),
                                                   'desc': ('CatDesc', str),
                                                   'plot': ('FieldNam', str),
                                                   'axis': ('LablAxis', str),
                                                   'scale': ('ScaleTyp', str),
                                                   'min_val': (
                                                       'ValidMin', float),
                                                   'max_val': (
                                                       'ValidMax', float),
                                                   'fill_val': (
                                                       'FillVal', float)},
                                           pandas_format=pandas_format)
    return data, mdata


# Recommended function
def list_remote_files(tag, inst_id, user=None, password=None):
    """Return a Pandas Series of every file for chosen remote data.

    This routine is intended to be used by pysat instrument modules supporting
    a particular NASA CDAWeb dataset.

    Parameters
    -----------
    tag : string or NoneType
        Denotes type of file to load.  Accepted types are <tag strings>.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    user : string or NoneType
        Username to be passed along to resource with relevant data.
        (default=None)
    password : string or NoneType
        User password to be passed along to resource with relevant data.
        (default=None)

    Note
    ----
    If defined, the expected return variable is a pandas.Series formatted for
    the Files class (pysat._files.Files) containing filenames and indexed by
    date and time

    """

    return



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------

import datetime as dt
import importlib
import netCDF4
import numpy as np
import os
import pandas as pds
from portalocker import Lock
import xarray as xr

import pysat


def scale_units(out_unit, in_unit):
    """ Determine the scaling factor between two units

    Parameters
    ----------
    out_unit : str
        Desired unit after scaling
    in_unit : str
        Unit to be scaled

    Returns
    -------
    unit_scale : float
        Scaling factor that will convert from in_units to out_units

    Note
    ----
    Accepted units include degrees ('deg', 'degree', 'degrees'),
    radians ('rad', 'radian', 'radians'),
    hours ('h', 'hr', 'hrs', 'hour', 'hours'), and lengths ('m', 'km', 'cm').
    Can convert between degrees, radians, and hours or different lengths.

    Example
    -------
    ::

        import numpy as np
        two_pi = 2.0 * np.pi
        scale = scale_units("deg", "RAD")
        two_pi *= scale
        two_pi # will show 360.0


    """

    if out_unit == in_unit:
        return 1.0

    accepted_units = {'deg': ['deg', 'degree', 'degrees'],
                      'rad': ['rad', 'radian', 'radians'],
                      'h': ['h', 'hr', 'hrs', 'hours'],
                      'm': ['m', 'km', 'cm'],
                      'm/s': ['m/s', 'cm/s', 'km/s', 'm s$^{-1}$',
                              'cm s$^{-1}$', 'km s$^{-1}$', 'm s-1', 'cm s-1',
                              'km s-1']}
    replace_str = {'/s': [' s$^{-1}$', ' s-1']}

    scales = {'deg': 180.0, 'rad': np.pi, 'h': 12.0,
              'm': 1.0, 'km': 0.001, 'cm': 100.0,
              'm/s': 1.0, 'cm/s': 100.0, 'km/s': 0.001}

    # Test input and determine transformation type
    out_key = out_unit.lower()
    in_key = in_unit.lower()
    for kk in accepted_units.keys():
        if out_key in accepted_units.keys() and in_key in accepted_units.keys():
            break

        if (out_key not in accepted_units.keys()
                and out_unit.lower() in accepted_units[kk]):
            out_key = kk
        if (in_key not in accepted_units.keys()
                and in_unit.lower() in accepted_units[kk]):
            in_key = kk

    if (out_key not in accepted_units.keys()
            and in_key not in accepted_units.keys()):
        raise ValueError(''.join(['Cannot scale {:s} and '.format(in_unit),
                                  '{:s}, unknown units'.format(out_unit)]))

    if out_key not in accepted_units.keys():
        raise ValueError('Unknown output unit {:}'.format(out_unit))

    if in_key not in accepted_units.keys():
        raise ValueError('Unknown input unit {:}'.format(in_unit))

    if out_key == 'm' or out_key == 'm/s' or in_key == 'm' or in_key == 'm/s':
        if in_key != out_key:
            raise ValueError('Cannot scale {:s} and {:s}'.format(out_unit,
                                                                 in_unit))
        # Recast units as keys for the scales dictionary and ensure that
        # the format is consistent
        rkey = ''
        for rr in replace_str.keys():
            if out_key.find(rr):
                rkey = rr

        out_key = out_unit.lower()
        in_key = in_unit.lower()

        if rkey in replace_str.keys():
            for rval in replace_str[rkey]:
                out_key = out_key.replace(rval, rkey)
                in_key = in_key.replace(rval, rkey)

    unit_scale = scales[out_key] / scales[in_key]

    return unit_scale


def load_netcdf4(fnames=None, strict_meta=False, file_format=None,
                 epoch_name='Epoch', pandas_format=True,
                 labels={'units': ('units', str), 'name': ('long_name', str),
                         'notes': ('notes', str), 'desc': ('desc', str),
                         'plot': ('plot_label', str), 'axis': ('axis', str),
                         'scale': ('scale', str),
                         'min_val': ('value_min', float),
                         'max_val': ('value_max', float),
                         'fill_val': ('fill', float)}):
    """Load netCDF-3/4 file produced by pysat.

    Parameters
    ----------
    fnames : string, array_like of strings, or NoneType
        Filename(s) to load, will fail if None (default=None)
    strict_meta : boolean
        Flag that checks if metadata across fnames is the same if True
        (default=False)
    file_format : string or NoneType
        file_format keyword passed to netCDF4 routine.  Expects one of
        'NETCDF3_CLASSIC', 'NETCDF3_64BIT', 'NETCDF4_CLASSIC', or 'NETCDF4'.
        If None, defaults to 'NETCDF4'. (default=None)
    epoch_name : string
        Data key for time variable (default='Epoch')
    pandas_format : bool
        Flag specifying if data is stored in a pandas DataFrame (True) or
        xarray Dataset (False). (default=False)
    labels : dict
        Dict where keys are the label attribute names and the values are tuples
        that have the label values and value types in that order.
        (default={'units': ('units', str), 'name': ('long_name', str),
        'notes': ('notes', str), 'desc': ('desc', str),
        'plot': ('plot_label', str), 'axis': ('axis', str),
        'scale': ('scale', str), 'min_val': ('value_min', float),
        'max_val': ('value_max', float), 'fill_val': ('fill', float)})

    Returns
    --------
    out : pandas.DataFrame
        DataFrame output
    meta : pysat.Meta
        Meta data

    Raises
    ------
    ValueError
        If kwargs that should be args are not set on instantiation.

    """

    if fnames is None:
        raise ValueError("Must supply a filename/list of filenames")
    if isinstance(fnames, str):
        fnames = [fnames]

    if file_format is None:
        file_format = 'NETCDF4'
    else:
        file_format = file_format.upper()

    saved_meta = None
    running_idx = 0
    running_store = []
    two_d_keys = []
    two_d_dims = []
    meta = pysat.Meta(labels=labels)

    if pandas_format:
        for fname in fnames:
            with netCDF4.Dataset(fname, mode='r', format=file_format) as data:
                # Build a dictionary with all global ncattrs and add those
                # attributes to a pysat meta object
                for ncattr in data.ncattrs():
                    if hasattr(meta, ncattr):
                        meta.__setattr__('{:}_'.format(ncattr),
                                         data.getncattr(ncattr))
                    else:
                        meta.__setattr__(ncattr, data.getncattr(ncattr))

                loaded_vars = {}
                for key in data.variables.keys():
                    # Load the metadata.  From here group unique dimensions and
                    # act accordingly, 1D, 2D, 3D
                    if len(data.variables[key].dimensions) == 1:
                        if pandas_format:
                            # Load 1D data variables, assuming basic
                            # time dimension
                            loaded_vars[key] = data.variables[key][:]

                        # Load up metadata
                        meta_dict = {}
                        for nc_key in data.variables[key].ncattrs():
                            meta_dict[nc_key] = \
                                data.variables[key].getncattr(nc_key)
                        meta[key] = meta_dict

                    if len(data.variables[key].dimensions) == 2:
                        # Part of dataframe within dataframe
                        two_d_keys.append(key)
                        two_d_dims.append(data.variables[key].dimensions)

                    if len(data.variables[key].dimensions) >= 3:
                        raise ValueError(' '.join(('pysat only supports 1D',
                                                   'and 2D data in pandas.',
                                                   'Please use xarray for',
                                                   'this data product.')))

                # We now have a list of keys that need to go into a dataframe,
                # could be more than one, collect unique dimensions for 2D keys
                for dim in set(two_d_dims):
                    # First or second dimension could be epoch. Use other
                    # dimension name as variable name
                    if dim[0] == epoch_name:
                        obj_key = dim[1]
                    elif dim[1] == epoch_name:
                        obj_key = dim[0]
                    else:
                        raise KeyError('Epoch not found!')
                    # collect variable names associated with dimension
                    idx_bool = [dim == i for i in two_d_dims]
                    idx, = np.where(np.array(idx_bool))
                    obj_var_keys = []
                    clean_var_keys = []
                    for i in idx:
                        obj_var_keys.append(two_d_keys[i])
                        clean_var_keys.append(
                            two_d_keys[i].split(obj_key + '_')[-1])

                    # Figure out how to index this data, it could provide its
                    # own index - or we may have to create simple integer based
                    # DataFrame access. If the dimension is stored as its own
                    # variable then use that info for index
                    if obj_key in obj_var_keys:
                        # string used to indentify dimension also in
                        # data.variables will be used as an index
                        index_key_name = obj_key
                        # if the object index uses UNIX time, process into
                        # datetime index
                        if data.variables[obj_key].getncattr(
                                meta.labels.name) == epoch_name:
                            # Name to be used in DataFrame index
                            index_name = epoch_name
                            time_index_flag = True
                        else:
                            time_index_flag = False
                            # Label to be used in DataFrame index
                            index_name = data.variables[obj_key].getncattr(
                                meta.labels.name)
                    else:
                        # Dimension is not itself a variable
                        index_key_name = None

                    # Iterate over the variables and grab metadata
                    dim_meta_data = pysat.Meta(labels=labels)

                    for key, clean_key in zip(obj_var_keys, clean_var_keys):
                        # store attributes in metadata, exept for dim name
                        meta_dict = {}
                        for nc_key in data.variables[key].ncattrs():
                            meta_dict[nc_key] = \
                                data.variables[key].getncattr(nc_key)
                        dim_meta_data[clean_key] = meta_dict

                    dim_meta_dict = {'meta': dim_meta_data}
                    if index_key_name is not None:
                        # Add top level meta
                        for nc_key in data.variables[obj_key].ncattrs():
                            dim_meta_dict[nc_key] = \
                                data.variables[obj_key].getncattr(nc_key)
                        meta[obj_key] = dim_meta_dict

                    # Iterate over all variables with this dimension
                    # data storage, whole shebang
                    loop_dict = {}

                    # List holds a series of slices, parsed from dict above
                    loop_list = []
                    for key, clean_key in zip(obj_var_keys, clean_var_keys):
                        # data
                        loop_dict[clean_key] = \
                            data.variables[key][:, :].flatten(order='C')

                    # Number of values in time
                    loop_lim = data.variables[obj_var_keys[0]].shape[0]

                    # Number of values per time
                    step = len(data.variables[obj_var_keys[0]][0, :])

                    # Check if there is an index we should use
                    if not (index_key_name is None):
                        # An index was found
                        time_var = loop_dict.pop(index_key_name)
                        if time_index_flag:
                            # Create datetime index from data
                            time_var = pds.to_datetime(1E6 * time_var)
                        new_index = time_var
                        new_index_name = index_name
                    else:
                        # Using integer indexing
                        new_index = np.arange((loop_lim * step),
                                              dtype=int) % step
                        new_index_name = 'index'

                    # Load all data into frame
                    if len(loop_dict.keys()) > 1:
                        loop_frame = pds.DataFrame(loop_dict,
                                                   columns=clean_var_keys)
                        if obj_key in loop_frame:
                            del loop_frame[obj_key]

                        # Break massive frame into bunch of smaller frames
                        for i in np.arange(loop_lim, dtype=int):
                            loop_list.append(loop_frame.iloc[(step * i):
                                                             (step * (i + 1)),
                                                             :])
                            loop_list[-1].index = new_index[(step * i):
                                                            (step * (i + 1))]
                            loop_list[-1].index.name = new_index_name
                    else:
                        loop_frame = pds.Series(loop_dict[clean_var_keys[0]],
                                                name=obj_var_keys[0])

                        # Break massive series into bunch of smaller series
                        for i in np.arange(loop_lim, dtype=int):
                            loop_list.append(loop_frame.iloc[(step * i):
                                                             (step * (i + 1))])
                            loop_list[-1].index = new_index[(step * i):
                                                            (step * (i + 1))]
                            loop_list[-1].index.name = new_index_name

                    # Add 2D object data, all based on a unique dimension within
                    # netCDF, to loaded data dictionary
                    loaded_vars[obj_key] = loop_list
                    del loop_list

                # Prepare dataframe index for this netcdf file
                time_var = loaded_vars.pop(epoch_name)

                # Convert from GPS seconds to seconds used in pandas (unix time,
                # no leap)
                # time_var = convert_gps_to_unix_seconds(time_var)
                loaded_vars[epoch_name] = pds.to_datetime(
                    (1E6 * time_var).astype(int))
                running_store.append(loaded_vars)
                running_idx += len(loaded_vars[epoch_name])

                if strict_meta:
                    if saved_meta is None:
                        saved_meta = meta.copy()
                    elif (meta != saved_meta):
                        raise ValueError(' '.join(('Metadata across filenames',
                                                   'is not the same.')))

        # Combine all of the data loaded across files together
        out = []
        for item in running_store:
            out.append(pds.DataFrame.from_records(item, index=epoch_name))
        out = pds.concat(out, axis=0)
    else:
        if len(fnames) == 1:
            out = xr.open_dataset(fnames[0])
        else:
            out = xr.open_mfdataset(fnames, combine='by_coords')
        for key in out.variables.keys():
            # Copy the variable attributes from the data object to the metadata
            meta_dict = {}
            for nc_key in out.variables[key].attrs.keys():
                meta_dict[nc_key] = out.variables[key].attrs[nc_key]
            meta[key] = meta_dict
            # Remove variable attributes from the data object
            out.variables[key].attrs = {}
        # Copy the file attributes from the data object to the metadata
        for out_attr in out.attrs.keys():
            if hasattr(meta, out_attr):
                set_attr = "".join([out_attr, "_"])
                meta.__setattr__(set_attr, out.attrs[out_attr])
            else:
                meta.__setattr__(out_attr, out.attrs[out_attr])
        # Remove attributes from the data object
        out.attrs = {}

    return out, meta


def fmt_output_in_cols(out_strs, ncols=3, max_num=6, lpad=None):
    """ Format a string with desired output values in columns

    Parameters
    ----------
    out_strs : array-like
        Array like object containing strings to print
    ncols : int
        Number of columns to print (default=3)
    max_num : int
        Maximum number of out_strs members to print.  Best display achieved if
        this number is divisable by 2 and ncols (default=6)
    lpad : int or NoneType
        Left padding or None to use length of longest string + 1 (default=None)

    Returns
    -------
    output : string
        String with desired data formatted in columns

    """
    output = ""

    # Ensure output strings are array-like
    out_strs = np.asarray(out_strs)
    if out_strs.shape == ():
        out_strs = np.array([out_strs])

    # If there are more data values than desired, keep the first and last
    out_len = len(out_strs)
    middle = -1
    if out_len > max_num:
        nhalf = int(max_num / 2)
        middle = nhalf // ncols
        if middle == 0:
            middle = 1
        nsel = [0] if nhalf == 0 else [i for i in range(nhalf)]
        nsel.extend([i for i in np.arange(out_len - nhalf, out_len)])
    else:
        nsel = np.arange(0, out_len)
    sel_len = len(nsel)

    # If desired, determine the left padding spacing
    if lpad is None:
        lpad = max([len(ostr) for ostr in out_strs[nsel]]) + 1

    # Print out the groups of variables in rows
    num = sel_len // ncols
    for i in range(num):
        # If data has been cut, indicate this with an ellipses row
        if i == middle:
            middle = -1
            output += "...".center(lpad * ncols) + '\n'

        # Print out data for each selected column in this row
        for j in range(ncols):
            output += out_strs[nsel][ncols * i + j].ljust(lpad)
        output += '\n'

    # Print out remaining variables one at a time on a single line
    extra_cols = sel_len - ncols * num
    if extra_cols > 0:
        for i in range(extra_cols):
            if middle >= 0:
                if i == 0 and num > 0:
                    output += "...".center(lpad * ncols) + '\n'
                elif num == 0 and i == nhalf:
                    output += "...".center(lpad if lpad > 4 else 4)
            output += out_strs[nsel][i + ncols * num].ljust(lpad)
        output += '\n'

    return output


def generate_instrument_list(inst_loc, user_info=None):
    """Iterate through and classify instruments in a given subpackage.


    Parameters
    ----------
    inst_loc : python subpackage
        The location of the instrument subpackage to test, e.g.,
        'pysat.instruments'
    user_info : dict or NoneType
        Nested dictionary with user and password info for instrument module
        name.  If None, no user or password is assumed. (default=None)
        EX: user_info = {'jro_isr': {'user': 'myname', 'password': 'email'}}

    Returns
    -------
    output : dict
        Dictionary with keys 'names', 'download', 'no_download' that contain
        lists with different information for each key:
        'names' - list of platform_name combinations
        'download' - dict containing 'inst_module', 'tag', and 'inst_id' for
        instruments with download routines
        'no_download' - dict containing 'inst_module', 'tag', and 'inst_id' for
        instruments without download routines

    Note
    ----
    This routine currently supports classification of instruments for unit
    tests both in the core package and in seperate instrument packages that
    use pysat.

    """

    instrument_names = inst_loc.__all__
    instrument_download = []
    instrument_no_download = []

    # Look through list of available instrument modules in the given location
    for inst_module in instrument_names:
        try:
            module = importlib.import_module(''.join(('.', inst_module)),
                                             package=inst_loc.__name__)
        except ImportError:
            # If this can't be imported, we can't pull out the info for the
            # download / no_download tests.  Leaving in basic tests for all
            # instruments, but skipping the rest.  The import error will be
            # caught as part of the pytest.mark.all_inst tests in InstTestClass
            pass
        else:
            # try to grab basic information about the module so we
            # can iterate over all of the options
            try:
                info = module._test_dates
            except AttributeError:
                # If a module does not have a test date, add it anyway for
                # other tests.  This will be caught later by
                # InstTestClass.test_instrument_test_dates
                info = {}
                info[''] = {'': dt.datetime(2009, 1, 1)}
                module._test_dates = info
            for inst_id in info.keys():
                for tag in info[inst_id].keys():
                    inst_dict = {'inst_module': module, 'tag': tag,
                                 'inst_id': inst_id}
                    # Add username and password info if needed
                    if user_info and inst_module in user_info:
                        inst_dict['user_info'] = user_info[inst_module]
                    # Initialize instrument so that pysat can generate skip
                    # flags where appropriate
                    inst = pysat.Instrument(inst_module=module,
                                            tag=tag,
                                            inst_id=inst_id,
                                            temporary_file_list=True)
                    travis_skip = ((os.environ.get('TRAVIS') == 'true')
                                   and not inst._test_download_travis)
                    if inst._test_download:
                        if not travis_skip:
                            instrument_download.append(inst_dict)
                    elif not inst._password_req:
                        # we don't want to test download for this combo
                        # But we do want to test the download warnings
                        # for instruments without a password requirement
                        instrument_no_download.append(inst_dict)

    output = {'names': instrument_names,
              'download': instrument_download,
              'no_download': instrument_no_download}

    return output


def available_instruments(inst_loc=None):
    """Obtain basic information about instruments in a given subpackage.

    Parameters
    ----------
    inst_loc : python subpackage or NoneType
        The location of the instrument subpackage (e.g., pysat.instruments)
        or None to list all registered instruments (default=None)

    Returns
    -------
    inst_info : dict
        Nested dictionary with 'platform', 'name', 'inst_module',
        'inst_ids_tags', 'inst_id', and 'tag' with the tag descriptions given
        as the value for each unique dictionary combination.

    """

    def get_inst_id_dict(inst_module_name):
        try:
            module = importlib.import_module(inst_module_name)
            inst_ids = {inst_id: {tag: module.tags[tag]
                                  for tag in module.inst_ids[inst_id]}
                        for inst_id in module.inst_ids.keys()}
        except ImportError as ierr:
            inst_ids = {'ERROR': {'ERROR': str(ierr)}}
        return inst_ids

    # Get user modules dictionary.
    user_modules = pysat.params['user_modules']

    if inst_loc is None:
        # Access the registered instruments
        inst_info = dict()

        # Cycle through each instrument platform and name to reshape the
        # dictionary and get the instrument tags and inst_ids
        for platform in user_modules.keys():
            inst_info[platform] = dict()
            for name in user_modules[platform].keys():
                inst_ids = get_inst_id_dict(user_modules[platform][name])
                inst_info[platform][name] = {'inst_module':
                                             user_modules[platform][name],
                                             'inst_ids_tags': inst_ids}
    else:
        # Access the instruments in the specified module
        inst_mods = inst_loc.__all__
        inst_info = dict()

        # Cycle through the available instrument modules
        for inst_mod in inst_mods:
            # Get the platform, name, and instrument module name
            platform = inst_mod.split('_')[0]
            name = '_'.join(inst_mod.split('_')[1:])
            mod_name = '.'.join([inst_loc.__name__, inst_mod])

            # Initialize the dictionary for this platform and name
            if platform not in inst_info.keys():
                inst_info[platform] = dict()

            # Finalize the dictionary
            inst_info[platform][name] = {
                'inst_module': mod_name,
                'inst_ids_tags': get_inst_id_dict(mod_name)}

    return inst_info


def display_available_instruments(inst_loc=None, show_inst_mod=None,
                                  show_platform_name=None):
    """Display basic information about instruments in a given subpackage.

    Parameters
    ----------
    inst_loc : python subpackage or NoneType
        The location of the instrument subpackage (e.g., pysat.instruments)
        or None to list all registered instruments (default=None)
    show_inst_mod : boolean or NoneType
        Displays the instrument module if True, does not include it if False,
        and reverts to standard display based on inst_loc type if None.
        (default=None)
    show_platform_name : boolean or NoneType
        Displays the platform and name if True, does not include it if False,
        and reverts to standard display based on inst_loc type if None.
        (default=None)

    Note
    ----
    Prints to standard out, a user-friendly interface for availabe_instruments.
    Defaults to including the instrument module and not the platform/name values
    if inst_loc is an instrument module and to including the platform/name
    values and not the instrument module if inst_loc is None (listing the
    registered instruments).

    """

    inst_info = available_instruments(inst_loc)

    if show_platform_name is None and inst_loc is None:
        show_platform_name = True
        plat_keys = sorted([platform for platform in inst_info.keys()])
    else:
        plat_keys = inst_info.keys()

    if show_inst_mod is None and inst_loc is not None:
        show_inst_mod = True

    if show_platform_name:
        header = "Platform   Name   "
    else:
        header = ""

    if show_inst_mod:
        header = "{:s}Instrument_Module".format(header)

    header = "{:s}      [Tag   Inst_ID]  Description".format(header)
    print(header)
    print("-" * len(header))
    for platform in plat_keys:
        for name in inst_info[platform].keys():
            mod_str = ""
            for inst_id in inst_info[platform][name]['inst_ids_tags'].keys():
                for tag in inst_info[platform][name][
                        'inst_ids_tags'][inst_id].keys():
                    if len(mod_str) == 0:
                        if show_platform_name:
                            mod_str = "".join([platform.__repr__(), " ",
                                               name.__repr__(), " "])
                        if show_inst_mod:
                            mod_str = "{:s}{:s}".format(
                                mod_str,
                                inst_info[platform][name]['inst_module'])
                    else:
                        mod_str = " " * len(mod_str)

                    print("".join([mod_str, " [", tag.__repr__(), " ",
                                   inst_id.__repr__(), "]  ",
                                   inst_info[platform][name]['inst_ids_tags'][
                                       inst_id][tag]]))
    return


class NetworkLock(Lock):
    def __init__(self, *args, **kwargs):
        """Lock manager compatible with networked file systems.

        Parameters
        ----------
        *args : list reference
            References a list of input arguments
        **kwargs : dict reference
            References a dict of input keyword argument

        Note
        ----
        See portalocker.utils.Lock for more details
        (:class:`portalocker.utils.Lock`)

        Example
        -------
        ::

            from pysat.utils import NetworkLock

            with NetworkLock(file_to_be_written, 'w') as locked_file:
                locked_file.write('content')

        """

        if 'timeout' in kwargs:
            timeout = kwargs['timeout']
            del kwargs['timeout']
        else:
            timeout = pysat.params['file_timeout']

        super(NetworkLock, self).__init__(timeout=timeout,
                                          *args, **kwargs)

    def release(self):
        """Releases the Lock so the file system

        From portalocker docs:
          On some networked filesystems it might be needed to force
          a `os.fsync()` before closing the file so it's
          actually written before another client reads the file.

        """

        self.fh.flush()
        os.fsync(self.fh.fileno())

        super(NetworkLock, self).release()



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------
"""
tests the registration of user-defined modules
"""

import copy
import numpy as np
import pytest
import sys

import pysat
from pysat.utils import registry
from pysat.tests.registration_test_class import TestWithRegistration


def ensure_updated_stored_modules(modules):
    """Ensure stored pysat.params['user_modules'] updated
    to include modules

    Parameters
    ----------
    modules : list
        List of tuples (module_string, platform, name)
        that will be checked against stored
        pysat.params['user_modules']

    """
    # make sure filesystem was updated
    saved_modules = registry.load_saved_modules()
    for module_name, platform, name in modules:
        assert platform in saved_modules
        assert name in saved_modules[platform]
        assert module_name in saved_modules[platform][name]


def ensure_live_registry_updated(modules):
    """Ensure current pysat.params['user_modules'] updated
    to include modules

    Parameters
    ----------
    modules : list
        List of tuples (module_string, platform, name)
        that will be checked against stored
        pysat.params['user_modules']

    """

    for module_name, platform, name in modules:
        # check that global registry was updated
        assert platform in pysat.params['user_modules']
        assert name in pysat.params['user_modules'][platform]
        assert module_name in pysat.params['user_modules'][platform][name]


class TestRegistration(TestWithRegistration):

    # Set setup/teardown to the class defaults
    setup = TestWithRegistration.setup
    teardown = TestWithRegistration.teardown

    def test_duplicate_registration_error(self):
        """Test register error for duplicate package"""

        # Register all modules at once
        registry.register(self.module_names)

        # Verify instantiation by platform/name combo
        for self.module_name, self.platform, self.name in self.modules:
            self.module_name = None  # Must be None to use platform/name combo
            self.verify_inst_instantiation()

        # Check that global registry was updated
        ensure_live_registry_updated(self.modules)

        # Verify update
        ensure_updated_stored_modules(self.modules)

        # Create a new package, the same as an existing one but with a
        # different module name
        mod_name = self.module_names[0]
        sys.modules['pysat_error.test_faux_module'] = sys.modules[mod_name]

        # register packages again, this should error
        with pytest.raises(ValueError):
            registry.register(['pysat_error.test_faux_module'])

        return

    def test_duplicate_registration_overwrite(self):
        """Test register error for duplicate package"""

        # Register all modules at once
        registry.register(self.module_names)

        # Verify instantiation by by platform/name combo
        for self.module_name, self.platform, self.name in self.modules:
            self.module_name = None  # Must be None to use platform/name combo
            self.verify_inst_instantiation()

        # Check that global registry was updated
        ensure_live_registry_updated(self.modules)

        # Verify update
        ensure_updated_stored_modules(self.modules)

        # Create a new package, the same as an existing one, but with a
        # different module name
        mod_name = self.module_names[0]
        sys.modules['pysat_error.test_faux_module'] = sys.modules[mod_name]

        # Register packages again
        registry.register(['pysat_error.test_faux_module'], overwrite=True)
        self.modules[0] = ('pysat_error.test_faux_module', 'platname1', 'name1')

        # Check that global registry was updated
        ensure_live_registry_updated(self.modules)

        # Verify update
        ensure_updated_stored_modules(self.modules)

        return

    def test_registration(self):
        """Test registering multiple instruments at once"""

        # Register all modules at once
        registry.register(self.module_names)

        # Verify instantiation by by platform/name combo
        for self.module_name, self.platform, self.name in self.modules:
            self.module_name = None  # Must be None to use platform/name combo
            self.verify_inst_instantiation()

        # Verify registration
        ensure_live_registry_updated(self.modules)

        # Verify stored update
        ensure_updated_stored_modules(self.modules)

        return

    def test_platform_removal(self):
        """Test removing entire platform at once"""

        # register all modules at once
        registry.register(self.module_names)

        # Verify instantiation by by platform/name combo
        for self.module_name, self.platform, self.name in self.modules:
            self.module_name = None  # Must be None to use platform/name combo
            self.verify_inst_instantiation()

        # verify registration
        ensure_live_registry_updated(self.modules)

        # verify stored update
        ensure_updated_stored_modules(self.modules)

        # remove them using only platform
        uplatforms = np.unique(self.platforms)
        registry.remove(uplatforms, [None] * len(uplatforms))

        return

    def test_platform_removal_single(self):
        """Test removing single platform at a time"""

        # Register all modules at once
        registry.register(self.module_names)

        # Verify instantiation by by platform/name combo
        for self.module_name, self.platform, self.name in self.modules:
            self.module_name = None  # Must be None to use platform/name combo
            self.verify_inst_instantiation()

        # Verify registration
        ensure_live_registry_updated(self.modules)

        # Verify stored update
        ensure_updated_stored_modules(self.modules)

        # Remove modules using only the platform. Doing this one by one ensures
        # more lines are tested  and other registered packages are still
        # present until their removal is performed.
        uplatforms, idx = np.unique(self.platforms, return_index=True)
        umodules = np.asarray(self.modules)[idx]
        for i, self.platform in enumerate(uplatforms):
            # Remove all Instruments on this platform and verify absence
            registry.remove([self.platform], [None])
            self.modules = [umodules[i]]
            self.ensure_not_in_stored_modules()

            # Test that other names still present
            if i < len(self.platforms) - 1:
                ensure_updated_stored_modules(umodules[i + 1:])

        return

    def test_platform_name_removal_single(self):
        """Test removing single platform/name at a time"""

        # Register all modules at once
        registry.register(self.module_names)

        # Verify instantiation by by platform/name combo
        for self.module_name, self.platform, self.name in self.modules:
            self.module_name = None  # Must be None to use platform/name combo
            self.verify_inst_instantiation()

        # Verify registration
        ensure_live_registry_updated(self.modules)

        # Verify stored update
        ensure_updated_stored_modules(self.modules)

        # Remove them using platform and name
        all_modules = list(self.modules)
        for i, self.platform in enumerate(self.platforms):
            # Set the name and module for removal
            self.name = self.names[i]
            self.modules = [all_modules[i]]

            # Remove by platform and name, verifying results
            registry.remove([self.platform], [self.name])
            self.ensure_not_in_stored_modules()

            # Ensure other names and platforms are still present
            if i < len(self.platforms) - 1:
                ensure_updated_stored_modules(self.modules[i + 1:])

        return

    def test_platform_removal_not_present(self):
        """Test error raised when removing module not present"""

        # Try to remove non-registered modules using only platform
        stored_modules = copy.deepcopy(pysat.params['user_modules'])
        registry.remove(['made_up_name'], [None])

        # Make sure nothing changed
        assert stored_modules == pysat.params['user_modules']

        return

    @pytest.mark.parametrize("par_plat, par_name",
                             [(['made_up_name', 'second'], ['made_up_name']),
                              (['made_up_name'], ['made_up_name', 'second']),
                              ([], ['made_up_name', 'second']),
                              ([], ['made_up_name'])])
    def test_platform_name_length_removal_error(self, par_plat, par_name):
        """Test error raised when platforms and names unequal lengths"""

        # Register all modules at once
        registry.register(self.module_names)

        # Raise error when removing non-existent Instruments
        with pytest.raises(ValueError):
            registry.remove(par_plat, par_name)

        return

    def test_module_registration_single(self):
        """Test registering a module containing an instrument"""

        # Register and verify package by module
        for self.module_name, self.platform, self.name in self.modules:
            root = self.module_name.split('.')[0]
            registry.register_by_module(sys.modules[root])
            self.module_name = None  # Test only works here by platform/name
            self.verify_inst_instantiation()

        # Check that global registry was updated
        ensure_live_registry_updated(self.modules)

        # Verify update
        ensure_updated_stored_modules(self.modules)

        return

    def test_module_registration_non_importable(self):
        """Test registering a non-existent module"""

        with pytest.raises(Exception):
            registry.register(['made.up.module'])

        return


class TestModuleRegistration():
    def setup(self):

        self.inst_module = pysat.instruments
        # package name
        pkg_name = self.inst_module.__name__

        # construct inputs similar to TestRegistration
        # to enable use of existing methods
        # almost fully general
        module_names = self.inst_module.__all__
        self.names = [snip.split('pysat')[-1][1:] for snip in module_names]
        self.platforms = ['pysat'] * len(self.names)
        module_strings = ['.'.join((pkg_name, name)) for name in
                          module_names]
        self.modules = [(mod, plat, nam) for mod, plat, nam in
                        zip(module_strings, self.platforms, self.names)]

        # remove any existing support which may be let over
        registry.remove(self.platforms, [None] * len(self.platforms))

        return

    def teardown(self):
        # clean up
        registry.remove(self.platforms, self.names)

        return

    def test_module_registration_multiple(self):
        """Test registering a module containing multiple instruments"""

        # register package by module
        registry.register_by_module(self.inst_module)
        # check that global registry was updated
        ensure_live_registry_updated(self.modules)
        # verify update on disk
        ensure_updated_stored_modules(self.modules)

        return



# -*- coding: utf-8 -*-
"""Supports the Neutral Atmosphere Composition Spectrometer (NACS) instrument
on Dynamics Explorer 2 (DE2).

From CDAWeb:

The Neutral Atmosphere Composition Spectrometer (NACS) was designed to obtain
in situ measurements of the neutral atmospheric composition and to study the
variations of the neutral atmosphere in response to energy coupled into it from
the magnetosphere.  Because temperature enhancements, large-scale circulation
cells, and wave propagation are produced by energy input (each of which
posseses a specific signature in composition variation), the measurements
permitted the study of the partition, flow, and deposition of energy from the
magnetosphere. Specifically, the investigation objective was to characterize
the composition of the neutral atmosphere with particular emphasis on
variability in constituent densities driven by interactions in the atmosphere,
ionosphere, and magnetosphere system. The quadrupole mass spectrometer used was
nearly identical to those flown on the AE-C, -D, and -E missions. The electron-
impact ion source was used in a closed mode. Atmospheric particles entered an
antechamber through a knife-edged orifice, where they were thermalized to the
instrument temperature. The ions with the selected charge-to-mass ratios had
stable trajectories through the hyperbolic electric field, exited the analyzer,
and entered the detection system. An off-axis beryllium-copper dynode
multiplier operating at a gain of 2.E6 provided an output pulse of electrons
for each ion arrival. The detector output had a pulse rate proportional to the
neutral density in the ion source of the selected mass. The instrument also
included two baffles that scanned across the input orifice for optional
measurement of the zonal and vertical components of the neutral wind. The mass
select system provided for 256 mass values between 0 and 51 atomic mass units
(u) or each 0.2 u. It was possible to call any one of these mass numbers into
each of eight 0.016-s intervals. This sequence was repeated each 0.128 s.

This data set includes daily files of the PI-provided DE-2 NACS 1-second data
and corresponding orbit parameters.  The data set was generated at NSSDC from
the original PI-provided data and software (SPTH-00010) and from the
orbit/attitude database and software that is part of the DE-2 UA data set
(SPIO-00174). The original NACS data were provided by the PI team in a highly
compressed VAX/VMS binary format on magnetic tapes. The data set covers the
whole DE-2 mission time period. Each data point is an average over the normally
8 measurements per second. Densities and relative errors are provided for
atomic oxygen (O), molecular nitrogen (N2), helium (He), atomic nitrogen (N),
and argon (Ar).  The data quality is generally quite good below 500 km, but
deteriorates towards higher altitudes as oxygen and molecular nitrogen approach
their background values (which could only be determined from infrequent
spinning orbits) and the count rate for Ar becomes very low. The difference
between minimum (background) and maximum count rate for atomic nitrogen
(estimated from mass 30) was so small that results are generally poor.  Data
were lost between 12 March 1982 and 31 March 1982 when the counter overflowed.


References
----------
G. R. Carrignan, B. P. Block, J. C. Maurer,  A. E. Hedin, C. A. Reber,
N. W. Spencer
The neutral mass spectrometer on Dynamics Explorer B
Space Sci. Instrum., v. 5, n. 4, p. 429, 1981.

Properties
----------
platform
    'de2'
name
    'nacs'
sat_id
    None Supported
tag
    None Supported

Authors
-------
J. Klenzing

"""

from __future__ import print_function
from __future__ import absolute_import

import functools

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'de2'
name = 'nacs'

tags = {'': '1 s cadence Neutral Atmosphere Composition Spectrometer data'}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(1983, 1, 1)}}

fname = 'de2_neutral1s_nacs_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}

# use the CDAWeb methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# use the default CDAWeb method
load = cdw.load

# support download routine
basic_tag = {'dir': '/pub/data/de/de2/neutral_gas_nacs/neutral1s_nacs_cdaweb',
             'remote_fname': '{year:4d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)

# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return PLATFORM/NAME data cleaned to the specified level

    Cleaning level is specified in inst.clean_level and pysat
    will accept user input for several strings. The clean_level is
    specified at instantiation of the Instrument object.

    'clean' All parameters should be good, suitable for statistical and
            case studies
    'dusty' All paramers should generally be good though same may
            not be great
    'dirty' There are data areas that have issues, data should be used
            with caution
    'none'  No cleaning applied, routine not called in this case.


    Parameters
    -----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    """

    return



"""
pysat.utils.time - date and time operations in pysat
=========================================

pysat.utils.time contains a number of functions used throughout
the pysat package, including interactions with datetime objects,
seasons, and calculation of solar local time
"""

import numpy as np
import pandas as pds
from pysat import datetime


def getyrdoy(date):
    """Return a tuple of year, day of year for a supplied datetime object.

    Parameters
    ----------
    date : datetime.datetime
        Datetime object

    Returns
    -------
    year : int
        Integer year
    doy : int
        Integer day of year

    """

    try:
        doy = date.toordinal() - datetime(date.year, 1, 1).toordinal() + 1
    except AttributeError:
        raise AttributeError("Must supply a pandas datetime object or " +
                             "equivalent")
    else:
        return date.year, doy


def parse_date(str_yr, str_mo, str_day, str_hr='0', str_min='0', str_sec='0',
               century=2000):
    """ Basic date parser for file reading

    Parameters
    ----------
    str_yr : string
        String containing the year (2 or 4 digits)
    str_mo : string
        String containing month digits
    str_day : string
        String containing day of month digits
    str_hr : string ('0')
        String containing the hour of day
    str_min : string ('0')
        String containing the minutes of hour
    str_sec : string ('0')
        String containing the seconds of minute
    century : int (2000)
        Century, only used if str_yr is a 2-digit year

    Returns
    -------
    out_date : pds.datetime
        Pandas datetime object

    """

    yr = int(str_yr) + century if len(str_yr) == 2 else int(str_yr)
    out_date = pds.datetime(yr, int(str_mo), int(str_day), int(str_hr),
                            int(str_min), int(str_sec))

    return out_date


def calc_freq(index):
    """ Determine the frequency for a time index

    Parameters
    ----------
    index : (array-like)
        Datetime list, array, or Index

    Returns
    -------
    freq : (str)
       Frequency string as described in Pandas Offset Aliases

    Notes
    -----
    Calculates the minimum time difference and sets that as the frequency.

    To reduce the amount of calculations done, the returned frequency is
    either in seconds (if no sub-second resolution is found) or nanoseconds.

    """

    # Test the length of the input
    if len(index) < 2:
        raise ValueError("insufficient data to calculate frequency")

    # Calculate the minimum temporal difference
    del_time = (np.array(index[1:]) - np.array(index[:-1])).min()

    # Convert minimum to seconds
    try:
        # First try as timedelta
        freq_sec = del_time.total_seconds()
    except AttributeError as err:
        # Now try as numpy.timedelta64
        if isinstance(del_time, np.timedelta64):
            freq_sec = float(del_time) * 1.0e-9
        else:
            raise AttributeError("Input should be times: {:}".format(err))

    # Format output frequency
    if np.floor(freq_sec) == freq_sec:
        # The frequency is on the order of seconds or greater
        freq = "{:.0f}S".format(freq_sec)
    else:
        # There are sub-seconds.  Go straigt to nanosec for best resoution
        freq = "{:.0f}N".format(freq_sec * 1.0e9)

    return freq


def season_date_range(start, stop, freq='D'):
    """
    Deprecated Function, will be removed in future version.

    .. deprecated:: 2.1.0
      `season_date_range` will be removed in pysat 3.0.0, this will be
      replaced by create_date_range



    """

    import warnings

    warnings.warn(' '.join(["utils.time.season_date_range is deprecated, use",
                            "utils.time.create_date_range instead"]),
                  DeprecationWarning, stacklevel=2)

    season = create_date_range(start, stop, freq=freq)

    return season


def create_date_range(start, stop, freq='D'):
    """
    Return array of datetime objects using input frequency from start to stop

    Supports single datetime object or list, tuple, ndarray of start and
    stop dates.

    freq codes correspond to pandas date_range codes, D daily, M monthly,
    S secondly

    """

    if hasattr(start, '__iter__'):
        # missing check for datetime
        season = pds.date_range(start[0], stop[0], freq=freq)
        for (sta, stp) in zip(start[1:], stop[1:]):
            season = season.append(pds.date_range(sta, stp, freq=freq))
    else:
        season = pds.date_range(start, stop, freq=freq)
    return season


def create_datetime_index(year=None, month=None, day=None, uts=None):
    """Create a timeseries index using supplied year, month, day, and ut in
    seconds.

    Parameters
    ----------
        year : array_like of ints
        month : array_like of ints or None
        day : array_like of ints
            for day (default) or day of year (use month=None)
        uts : array_like of floats

    Returns
    -------
        Pandas timeseries index.

    Note
    ----
    Leap seconds have no meaning here.

    """

    # need a timeseries index for storing satellite data in pandas but
    # creating a datetime object for everything is too slow
    # so I calculate the number of nanoseconds elapsed since first sample,
    # and create timeseries index from that.
    # Factor of 20 improvement compared to previous method,
    # which itself was an order of magnitude faster than datetime.

    # get list of unique year, and month
    if not hasattr(year, '__iter__'):
        raise ValueError('Must provide an iterable for all inputs.')
    if len(year) == 0:
        raise ValueError('Length of array must be larger than 0.')
    year = year.astype(int)
    if month is None:
        month = np.ones(len(year), dtype=int)
    else:
        month = month.astype(int)

    if uts is None:
        uts = np.zeros(len(year))
    if day is None:
        day = np.ones(len(year))
    day = day.astype(int)
    # track changes in seconds
    uts_del = uts.copy().astype(float)
    # determine where there are changes in year and month that need to be
    # accounted for
    _, idx = np.unique(year*100.+month, return_index=True)
    # create another index array for faster algorithm below
    idx2 = np.hstack((idx, len(year) + 1))
    # computes UTC seconds offset for each unique set of year and month
    for _idx, _idx2 in zip(idx[1:], idx2[2:]):
        temp = (datetime(year[_idx], month[_idx], 1) -
                datetime(year[0], month[0], 1))
        uts_del[_idx:_idx2] += temp.total_seconds()

    # add in UTC seconds for days, ignores existence of leap seconds
    uts_del += (day - 1) * 86400
    # add in seconds since unix epoch to first day
    uts_del += (datetime(year[0], month[0], 1) -
                datetime(1970, 1, 1)).total_seconds()
    # going to use routine that defaults to nanseconds for epoch
    uts_del *= 1E9
    return pds.to_datetime(uts_del)



# -*- coding: utf-8 -*-
"""
Ion Velocity Meter (IVM) support for the NASA/INPE SPORT CubeSat.

This mission is still in development. This routine is here to help
with the development of code associated with SPORT and the IVM.

"""

import functools
import logging
import warnings

import pysat
from pysat.instruments.methods import general as mm_gen

logger = logging.getLogger(__name__)


# pysat required parameters
platform = 'sport'
name = 'ivm'
# dictionary of data 'tags' and corresponding description
tags = {'': 'Level-2 IVM Files',
        'L1': 'Level-1 IVM Files',
        'L0': 'Level-0 IVM Files'}
# dictionary of satellite IDs, list of corresponding tags
# only one satellite in this case
sat_ids = {'': ['']}
# good day to download test data for. Downloads aren't currently supported
_test_dates = {'': {'': pysat.datetime(2019, 1, 1)}}

prefix = 'SPORT_{tag}_IVM_'
format_str = ''.join(('{year:04d}-{month:02d}-{day:02d}',
                      '_v{version:02d}r{revision:04d}.NC'))
supported_tags = {'': {'': ''.join((prefix.format(tag='L2'), format_str)),
                       'L1': ''.join((prefix.format(tag='L1'), format_str)),
                       'L0': ''.join((prefix.format(tag='L0'), format_str))}}
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    """

    logger.info(' '.join(("Mission acknowledgements and data restrictions will",
                          "be printed here when available.")))

    pass


def load(fnames, tag=None, sat_id=None, **kwargs):
    """Loads SPORT IVM data using pysat.utils.load_netcdf4.

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    fnames : array-like
        iterable of filename strings, full path, to data files to be loaded.
        This input is nominally provided by pysat itself.
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    sat_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    **kwargs : extra keywords
        Passthrough for additional keyword arguments specified when
        instantiating an Instrument object. These additional keywords
        are passed through to this routine by pysat.

    Returns
    -------
    data, metadata
        Data and Metadata are formatted for pysat. Data is a pandas
        DataFrame while metadata is a pysat.Meta instance.

    Note
    ----
    Any additional keyword arguments passed to pysat.Instrument
    upon instantiation are passed along to this routine and through
    to the load_netcdf4 call.

    Examples
    --------
    ::

        inst = pysat.Instrument('sport', 'ivm')
        inst.load(2019, 1)

    """

    return pysat.utils.load_netcdf4(fnames, **kwargs)


def download(date_array, tag, sat_id, data_path=None, user=None,
             password=None):
    """Downloads data for SPORT IVM, once SPORT is operational and in orbit.

    This routine is invoked by pysat and is not intended for direct use by
    the end user.

    Parameters
    ----------
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need
        not be contiguous.
    tag : string ('')
        Tag identifier used for particular dataset. This input is provided by
        pysat.
    sat_id : string  ('')
        Satellite ID string identifier used for particular dataset. This input
        is provided by pysat.
    data_path : string (None)
        Path to directory to download data to.
    user : string (None)
        User string input used for download. Provided by user and passed via
        pysat. If an account is required for dowloads this routine here must
        error if user not supplied.
    password : string (None)
        Password for data download.

    """

    warnings.warn('Downloads are not currently supported - not launched yet!')

    pass



# -*- coding: utf-8 -*-
"""
Supports loading data from files generated using TIEGCM
(Thermosphere Ionosphere Electrodynamics General Circulation Model) model.
TIEGCM file is a netCDF file with multiple dimensions for some variables.

Properties
----------
platform
    'ucar'
name
    'tiegcm'
tag
    None supported
sat_id
    None supported

"""

# python 2/3 comptability
from __future__ import print_function
from __future__ import absolute_import
import logging
import warnings

import xarray as xr
import pysat

logger = logging.getLogger(__name__)

# the platform and name strings associated with this instrument
# need to be defined at the top level
# these attributes will be copied over to the Instrument object by pysat
# the strings used here should also be used to name this file
# platform_name.py
platform = 'ucar'
name = 'tiegcm'

# dictionary of data 'tags' and corresponding description
tags = {'': 'UCAR TIE-GCM file'}
# dictionary of satellite IDs, list of corresponding tags for each sat_ids
# example
# sat_ids = {'a':['L1', 'L0'], 'b':['L1', 'L2'], 'c':['L1', 'L3']}
sat_ids = {'': ['']}
# good day to download test data for. Downloads aren't currently supported!
# format is outer dictionary has sat_id as the key
# each sat_id has a dictionary of test dates keyed by tag string
_test_dates = {'': {'': pysat.datetime(2019, 1, 1)}}

# specify using xarray (not using pandas)
pandas_format = False


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Parameters
    ----------
    self : pysat.Instrument
        This object

    """

    logger.info("Mission acknowledgements and data restrictions will be printed " +
          "here when available.")
    return


def load(fnames, tag=None, sat_id=None, **kwargs):
    """Loads TIEGCM data using xarray.

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    fnames : array-like
        iterable of filename strings, full path, to data files to be loaded.
        This input is nominally provided by pysat itself.
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. (default='')
    sat_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. (default='')
    **kwargs : extra keywords
        Passthrough for additional keyword arguments specified when
        instantiating an Instrument object. These additional keywords
        are passed through to this routine by pysat.

    Returns
    -------
    data, metadata
        Data and Metadata are formatted for pysat. Data is an xarray
        DataSet while metadata is a pysat.Meta instance.

    Note
    ----
    Any additional keyword arguments passed to pysat.Instrument
    upon instantiation are passed along to this routine.

    Examples
    --------
    ::

        inst = pysat.Instrument('ucar', 'tiegcm')
        inst.load(2019,1)

    """

    # load data
    data = xr.open_dataset(fnames[0])
    # move attributes to the Meta object
    # these attributes will be trasnferred to the Instrument object
    # automatically by pysat
    meta = pysat.Meta()
    for attr in data.attrs:
        setattr(meta, attr[0], attr[1])
    data.attrs = []

    # fill Meta object with variable information
    for key in data.variables.keys():
        attrs = data.variables[key].attrs
        meta[key] = attrs

    # move misc parameters from xarray to the Instrument object via Meta
    # doing this after the meta ensures all metadata is still kept
    # even for moved variables
    meta.p0 = data['p0']
    meta.p0_model = data['p0_model']
    meta.grav = data['grav']
    meta.mag = data['mag']
    meta.timestep = data['timestep']
    # remove these variables from xarray
    data = data.drop(['p0', 'p0_model', 'grav', 'mag', 'timestep'])

    return data, meta


def list_files(tag=None, sat_id=None, data_path=None, format_str=None):
    """Produce a list of files corresponding to UCAR TIEGCM.

    This routine is invoked by pysat and is not intended for direct
    use by the end user. Arguments are provided by pysat.

    Multiple data levels may be supported via the 'tag' input string.

    Parameters
    ----------
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. (default='')
    sat_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. (default='')
    data_path : string
        Full path to directory containing files to be loaded. This
        is provided by pysat. The user may specify their own data path
        at Instrument instantiation and it will appear here. (default=None)
    format_str : string
        String template used to parse the datasets filenames. If a user
        supplies a template string at Instrument instantiation
        then it will appear here, otherwise defaults to None. (default=None)

    Returns
    -------
    pandas.Series
        Series of filename strings, including the path, indexed by datetime.

    Examples
    --------
    ::

        If a filename is SPORT_L2_IVM_2019-01-01_v01r0000.NC then the template
        is 'SPORT_L2_IVM_{year:04d}-{month:02d}-{day:02d}_' +
        'v{version:02d}r{revision:04d}.NC'

    Note
    ----
    The returned Series should not have any duplicate datetimes. If there are
    multiple versions of a file the most recent version should be kept and the
    rest discarded. This routine uses the pysat.Files.from_os constructor, thus
    the returned files are up to pysat specifications.

    """

    if format_str is None:
        # default file naming
        format_str = 'tiegcm_icon_merg2.0_totTgcm.s_{day:03d}_{year:4d}.nc'

    return pysat.Files.from_os(data_path=data_path, format_str=format_str)


def download(date_array, tag, sat_id, data_path=None, user=None, password=None,
             **kwargs):
    """Placeholder for UCAR TIEGCM downloads. Doesn't do anything.

    This routine is invoked by pysat and is not intended for direct use by
    the end user.

    Parameters
    ----------
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need not
        be contiguous.
    tag : string
        Tag identifier used for particular dataset. This input is provided by
        pysat.
    sat_id : string
        Satellite ID string identifier used for particular dataset. This input
        is provided by pysat.
    data_path : string
        Path to directory to download data to. (default=None)
    user : string
        User string input used for download. Provided by user and passed via
        pysat. If an account is required for dowloads this routine here must
        error if user not supplied. (default=None)
    password : string
        Password for data download. (default=None)
    **kwargs : dict
        Additional keywords supplied by user when invoking the download
        routine attached to a pysat.Instrument object are passed to this
        routine via kwargs.

    """

    warnings.warn('Not implemented in this version.')
    return



# -*- coding: utf-8 -*-
"""
Produces fake instrument data for testing.
"""

import datetime as dt
import functools
import numpy as np

import pandas as pds

import pysat
from pysat.instruments.methods import testing as mm_test

logger = pysat.logger

platform = 'pysat'
name = 'testing2d'
tags = {'': 'Regular testing data set'}
inst_ids = {'': ['']}
_test_dates = {'': {'': dt.datetime(2009, 1, 1)}}


# Init method
init = mm_test.init


# Clean method
clean = mm_test.clean


# Optional method, preprocess
preprocess = mm_test.preprocess


def load(fnames, tag=None, inst_id=None, malformed_index=False,
         num_samples=None, test_load_kwarg=None):
    """ Loads the test files

    Parameters
    ----------
    fnames : list
        List of filenames
    tag : str or NoneType
        Instrument tag (accepts '')
    inst_id : str or NoneType
        Instrument satellite ID (accepts '')
    malformed_index : bool
        If True, the time index will be non-unique and non-monotonic.
        (default=False)
    num_samples : int
        Number of samples
    test_load_kwarg : any or NoneType
        Testing keyword (default=None)

    Returns
    -------
    data : pds.DataFrame
        Testing data
    meta : pysat.Meta
        Testing metadata

    """

    # Support keyword testing
    logger.info(''.join(('test_load_kwarg = ', str(test_load_kwarg))))

    # create an artifical satellite data set
    iperiod = mm_test.define_period()
    drange = mm_test.define_range()
    if num_samples is None:
        # Default to 1 day at a frequency of 100S
        num_samples = 864

    # Using 100s frequency for compatibility with seasonal analysis unit tests
    uts, index, dates = mm_test.generate_times(fnames, num_samples,
                                               freq='100S')
    # seed DataFrame with UT array
    data = pds.DataFrame(np.mod(uts, 86400.), columns=['uts'])

    # need to create simple orbits here. Have start of first orbit
    # at 2009,1, 0 UT. 14.84 orbits per day
    # figure out how far in time from the root start
    # use that info to create a signal that is continuous from that start
    # going to presume there are 5820 seconds per orbit (97 minute period)
    time_delta = dates[0] - dt.datetime(2009, 1, 1)
    # mlt runs 0-24 each orbit.
    data['mlt'] = mm_test.generate_fake_data(time_delta.total_seconds(), uts,
                                             period=iperiod['lt'],
                                             data_range=drange['lt'])
    # do slt, 20 second offset from mlt
    data['slt'] = mm_test.generate_fake_data(time_delta.total_seconds() + 20,
                                             uts, period=iperiod['lt'],
                                             data_range=drange['lt'])
    # create a fake longitude, resets every 6240 seconds
    # sat moves at 360/5820 deg/s, Earth rotates at 360/86400, takes extra time
    # to go around full longitude
    data['longitude'] = mm_test.generate_fake_data(time_delta.total_seconds(),
                                                   uts, period=iperiod['lon'],
                                                   data_range=drange['lon'])
    # create latitude signal for testing polar orbits
    angle = mm_test.generate_fake_data(time_delta.total_seconds(),
                                       uts, period=iperiod['angle'],
                                       data_range=drange['angle'])
    data['latitude'] = 90.0 * np.cos(angle)

    # create constant altitude at 400 km
    alt0 = 400.0
    data['altitude'] = alt0 * np.ones(data['latitude'].shape)

    if malformed_index:
        index = index.tolist()
        # nonmonotonic
        index[0:3], index[3:6] = index[3:6], index[0:3]
        # non unique
        index[6:9] = [index[6]] * 3

    data.index = index
    data.index.name = 'epoch'
    # higher rate time signal (for scalar >= 2)
    # this time signal used for 2D profiles associated with each time in main
    # DataFrame
    num_profiles = 50 if num_samples >= 50 else num_samples
    end_date = dates[0] + dt.timedelta(seconds=2 * num_profiles - 1)
    high_rate_template = pds.date_range(dates[0], end_date, freq='2S')

    # create a few simulated profiles
    # DataFrame at each time with mixed variables
    profiles = []
    # DataFrame at each time with numeric variables only
    alt_profiles = []
    # Serie at each time, numeric data only
    series_profiles = []
    # frame indexed by date times
    frame = pds.DataFrame({'density': data.loc[data.index[0:num_profiles],
                                               'mlt'].values.copy(),
                           'dummy_str': ['test'] * num_profiles,
                           'dummy_ustr': [u'test'] * num_profiles},
                          index=data.index[0:num_profiles],
                          columns=['density', 'dummy_str', 'dummy_ustr'])
    # frame indexed by float
    dd = np.arange(num_profiles) * 1.2
    ff = np.arange(num_profiles) / num_profiles
    ii = np.arange(num_profiles) * 0.5
    frame_alt = pds.DataFrame({'density': dd, 'fraction': ff},
                              index=ii,
                              columns=['density', 'fraction'])
    # series version of storage
    series_alt = pds.Series(dd, index=ii, name='series_profiles')

    for time in data.index:
        frame.index = high_rate_template + (time - data.index[0])
        profiles.append(frame)
        alt_profiles.append(frame_alt)
        series_profiles.append(series_alt)
    # store multiple data types into main frame
    data['profiles'] = pds.Series(profiles, index=data.index)
    data['alt_profiles'] = pds.Series(alt_profiles, index=data.index)
    data['series_profiles'] = pds.Series(series_profiles, index=data.index)

    # create very limited metadata
    meta = pysat.Meta()
    meta['uts'] = {'units': 's', 'long_name': 'Universal Time'}
    meta['mlt'] = {'units': 'hours', 'long_name': 'Magnetic Local Time'}
    meta['slt'] = {'units': 'hours', 'long_name': 'Solar Local Time'}
    meta['longitude'] = {'units': 'degrees', 'long_name': 'Longitude'}
    meta['latitude'] = {'units': 'degrees', 'long_name': 'Latitude'}
    meta['altitude'] = {'units': 'km', 'long_name': 'Altitude'}
    series_profile_meta = pysat.Meta()
    series_profile_meta['series_profiles'] = {'long_name': 'series'}
    meta['series_profiles'] = {'meta': series_profile_meta,
                               'long_name': 'series'}
    profile_meta = pysat.Meta()
    profile_meta['density'] = {'long_name': 'profiles'}
    profile_meta['dummy_str'] = {'long_name': 'profiles'}
    profile_meta['dummy_ustr'] = {'long_name': 'profiles'}
    meta['profiles'] = {'meta': profile_meta, 'long_name': 'profiles'}
    alt_profile_meta = pysat.Meta()
    alt_profile_meta['density'] = {'long_name': 'profiles'}
    alt_profile_meta['fraction'] = {'long_name': 'profiles'}
    meta['alt_profiles'] = {'meta': alt_profile_meta, 'long_name': 'profiles'}

    return data, meta


list_files = functools.partial(mm_test.list_files, test_dates=_test_dates)
list_remote_files = functools.partial(mm_test.list_remote_files,
                                      test_dates=_test_dates)
download = functools.partial(mm_test.download)



# -*- coding: utf-8 -*-
"""Supports the Wind and Temperature Spectrometer (WATS) instrument on
Dynamics Explorer 2 (DE2).

From CDAWeb:

The Wind and Temperature Spectrometer (WATS) measured the in situ neutral
winds, the neutral particle temperatures, and the concentrations of selected
gases. The objective of this investigation was to study the interrelationships
among the winds, temperatures, plasma drift, electric fields, and other
properties of the thermosphere that were measured by this and other instruments
on the spacecraft. Knowledge of how these properties are interrelated
contributed to an understanding of the consequences of the acceleration of
neutral particles by the ions in the ionosphere, the acceleration of ions by
neutrals creating electric fields, and the related energy transfer between the
ionosphere and the magnetosphere. Three components of the wind, one normal to
the satellite velocity vector in the horizontal plane, one vertical, and one in
the satellite direction were measured. A retarding potential quadrupole mass
spectrometer, coupled to the atmosphere through a precisely orificed
antechamber, was used. It was operated in either of two modes: one employed the
retarding capability and the other used the ion source as a conventional
nonretarding source. Two scanning baffles were used in front of the mass
spectrometer: one moved vertically and the other moved horizontally. The
magnitudes of the horizontal and vertical components of the wind normal to the
spacecraft velocity vector were computed from measurements of the angular
relationship between the neutral particle stream and the sensor. The component
of the total stream velocity in the satellite direction was measured directly
by the spectrometer system through determination of the required retarding
potential.  At altitudes too high for neutral species measurements, the planned
operation required the instrument to measure the thermal ion species only.  A
series of four sequentially occurring slots --each a 2-s long measurement
interval-- was adapted for the basic measurement format of the instrument.
Different functions were commanded into these slots in any combination, one per
measurement interval. Thus the time resolution can be 2, 4, 6, or 8 seconds.
Further details are found in This data set consists of the high-resolution data
of the Dynamics Explorer 2 Wind and Temperature Spectrometer (WATS) experiment.
The files contain the neutral density, temperature and horizontal (zonal) wind
velocity, and orbital parameters in ASCII format. The time resolution is
typically 2 seconds. Data are given as daily files (typically a few 100 Kbytes
each). PI-provided software (WATSCOR) was used to correct the binary data set.
NSSDC-developed software was used to add the orbit parameters, to convert the
binary into ASCII format and to combine the (PI-provided) orbital files into
daily files. For more on DE-2, WATS, and the binary data, see the
WATS_VOLDESC_SFDU_DE.DOC and WATS_FORMAT_SFDU_DE.DOC files. More information
about the processing done at NSSDC is given in WATS_NSSDC_PRO_DE.DOC.


References
----------
N. W. Spencer, L. E. Wharton, H. B. Niemann, A. E. Hedin, G. R. Carrignan,
J. C. Maurer
The Dynamics Explorer Wind and Temperature Spectrometer
Space Sci. Instrum., v. 5, n. 4, p. 417, 1981.

Properties
----------
platform
    'de2'
name
    'wats'
sat_id
    None Supported
tag
    None Supported

Authors
-------
J. Klenzing

"""

from __future__ import print_function
from __future__ import absolute_import

import functools

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'de2'
name = 'wats'

tags = {'': '2 s cadence Wind and Temperature Spectrometer data'}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(1983, 1, 1)}}

fname = 'de2_wind2s_wats_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}

# use the CDAWeb methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# use the default CDAWeb method
load = cdw.load

# support download routine
basic_tag = {'dir': '/pub/data/de/de2/neutral_gas_wats/wind2s_wats_cdaweb',
             'remote_fname': '{year:4d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)

# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return PLATFORM/NAME data cleaned to the specified level

    Cleaning level is specified in inst.clean_level and pysat
    will accept user input for several strings. The clean_level is
    specified at instantiation of the Instrument object.

    'clean' All parameters should be good, suitable for statistical and
            case studies
    'dusty' All paramers should generally be good though same may
            not be great
    'dirty' There are data areas that have issues, data should be used
            with caution
    'none'  No cleaning applied, routine not called in this case.


    Parameters
    -----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    """

    return



import pysat

"""
Creates a constellation for testing difference with two small identical
test instruments
"""

inst1 = pysat.Instrument('pysat', 'testing',
                         clean_level='clean',
                         sat_id='6000')
inst2 = pysat.Instrument('pysat', 'testing',
                         clean_level='clean',
                         sat_id='6000')
instruments = [inst1, inst2]



from __future__ import print_function
from __future__ import absolute_import

"""Occurrence probability routines, daily or by orbit.

Routines calculate the occurrence of an event greater than a supplied gate
occuring at least once per day, or once per orbit. The probability is
calculated as the (number of times with at least one hit in bin)/(number
of times in the bin).The data used to determine the occurrence must be 1D.
If a property of a 2D or higher dataset is needed attach a custom function
that performs the check and returns a 1D Series.

Note
----
The included routines use the bounds attached to the supplied instrument
object as the season of interest.

"""

import numpy as np
import warnings


def daily2D(inst, bin1, label1, bin2, label2, data_label, gate,
            returnBins=False):
    """2D Daily Occurrence Probability of data_label > gate over a season.

    .. deprecated:: 2.2.0
      `daily2D` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    If data_label is greater than gate at least once per day,
    then a 100% occurrence probability results.Season delineated by the bounds
    attached to Instrument object.
    Prob = (# of times with at least one hit)/(# of times in bin)

    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min, max, number of bins]
    labelx: string
        name for data product for binx
    data_label: list of strings
        identifies data product(s) to calculate occurrence probability
        e.g. inst[data_label]
    gate: list of values
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of days with any
        data; 'bin_x' and 'bin_y' are also returned if requested. Note that
        arrays are organized for direct plotting, y values along rows, x along
        columns.

    Note
    ----
    Season delineated by the bounds attached to Instrument object.

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    return _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate,
                         by_orbit=False, returnBins=returnBins)


def by_orbit2D(inst, bin1, label1, bin2, label2, data_label, gate,
               returnBins=False):
    """2D Occurrence Probability of data_label orbit-by-orbit over a season.

    .. deprecated:: 2.2.0
      `by_orbit2D` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    If data_label is greater than gate atleast once per orbit, then a
    100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.
    Prob = (# of times with at least one hit)/(# of times in bin)

    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min value, max value, number of bins]
    labelx: string
        identifies data product for binx
    data_label: list of strings
        identifies data product(s) to calculate occurrence probability
    gate: list of values
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of orbits with
        any data; 'bin_x' and 'bin_y' are also returned if requested. Note that
        arrays are organized for direct plotting, y values along rows, x along
        columns.

    Note
    ----
    Season delineated by the bounds attached to Instrument object.

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    return _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate,
                         by_orbit=True, returnBins=returnBins)


def _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate,
                  by_orbit=False, returnBins=False):
    if not hasattr(data_label, '__iter__'):
        raise ValueError('Data label must be list-like group of variable ' +
                         'names.')
    if not hasattr(gate, '__iter__'):
        raise ValueError('Gate levels must be list-like group of variable ' +
                         'names.')
    if len(gate) != len(data_label):
        raise ValueError('Must have a gate value for each data_label')

    # create bins
    binx = np.linspace(bin1[0], bin1[1], bin1[2] + 1)
    biny = np.linspace(bin2[0], bin2[1], bin2[2] + 1)

    numx = len(binx) - 1
    numy = len(biny) - 1
    numz = len(data_label)
    arrx = np.arange(numx)
    arry = np.arange(numy)
    arrz = np.arange(numz)

    # create arrays to store all values
    total = np.zeros((numz, numy, numx))
    hits = np.zeros((numz, numy, numx))
    if by_orbit:
        inst.load(date=inst.bounds[0][0])
        iterator = inst.orbits
    else:
        iterator = inst

    for i, inst in enumerate(iterator):
        if len(inst.data) != 0:
            xind = np.digitize(inst.data[label1], binx) - 1
            for xi in arrx:
                xindex, = np.where(xind == xi)
                if len(xindex) > 0:
                    yData = inst.data.iloc[xindex]
                    yind = np.digitize(yData[label2], biny) - 1
                    for yj in arry:
                        yindex, = np.where(yind == yj)
                        if len(yindex) > 0:
                            # iterate over the different data_labels
                            for zk in arrz:
                                indlab = yData.columns.get_loc(data_label[zk])
                                zdata = yData.iloc[yindex, indlab]
                                if np.any(np.isfinite(zdata)):
                                    total[zk, yj, xi] += 1.
                                    if np.any(zdata > gate[zk]):
                                        hits[zk, yj, xi] += 1.

    # all of the loading and storing data is done
    # get probability
    prob = hits/total
    # make nicer dictionary output
    output = {}
    for i, label in enumerate(data_label):
        output[label] = {'prob': prob[i, :, :], 'count': total[i, :, :]}
        if returnBins:
            output[label]['bin_x'] = binx
            output[label]['bin_y'] = biny
    # clean up
    del iterator
    return output


def daily3D(inst, bin1, label1, bin2, label2, bin3, label3,
            data_label, gate, returnBins=False):
    """3D Daily Occurrence Probability of data_label > gate over a season.

    .. deprecated:: 2.2.0
      `daily3D` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    If data_label is greater than gate atleast once per day,
    then a 100% occurrence probability results. Season delineated by
    the bounds attached to Instrument object.
    Prob = (# of times with at least one hit)/(# of times in bin)

    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min, max, number of bins]
    labelx: string
        name for data product for binx
    data_label: list of strings
        identifies data product(s) to calculate occurrence probability
    gate: list of values
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of days with any
        data; 'bin_x', 'bin_y', and 'bin_z' are also returned if requested.
        Note that arrays are organized for direct plotting, z,y,x.

    Note
    ----
    Season delineated by the bounds attached to Instrument object.

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    return _occurrence3D(inst, bin1, label1, bin2, label2, bin3, label3,
                         data_label, gate, returnBins=returnBins,
                         by_orbit=False)


def by_orbit3D(inst, bin1, label1, bin2, label2, bin3, label3,
               data_label, gate, returnBins=False):
    """3D Occurrence Probability of data_label orbit-by-orbit over a season.

    .. deprecated:: 2.2.0
      `by_orbit3D` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    If data_label is greater than gate atleast once per orbit, then a
    100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.
    Prob = (# of times with at least one hit)/(# of times in bin)

    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min value, max value, number of bins]
    labelx: string
        identifies data product for binx
    data_label: list of strings
        identifies data product(s) to calculate occurrence probability
    gate: list of values
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of orbits with
        any data; 'bin_x', 'bin_y', and 'bin_z' are also returned if requested.
        Note that arrays are organized for direct plotting, z,y,x.

    Note
    ----
    Season delineated by the bounds attached to Instrument object.

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    return _occurrence3D(inst, bin1, label1, bin2, label2, bin3, label3,
                         data_label, gate, returnBins=returnBins,
                         by_orbit=True)


def _occurrence3D(inst, bin1, label1, bin2, label2, bin3, label3,
                  data_label, gate, returnBins=False, by_orbit=False):

    if not hasattr(data_label, '__iter__'):
        raise ValueError('Data label must be list-like group of variable ' +
                         'names.')
    if not hasattr(gate, '__iter__'):
        raise ValueError('Gate levels must be list-like group of variable ' +
                         'names.')
    if len(gate) != len(data_label):
        raise ValueError('Must have a gate value for each data_label')

    # create bins
    binx = np.linspace(bin1[0], bin1[1], bin1[2] + 1)
    biny = np.linspace(bin2[0], bin2[1], bin2[2] + 1)
    binz = np.linspace(bin3[0], bin3[1], bin3[2] + 1)

    numx = len(binx) - 1
    numy = len(biny) - 1
    numz = len(binz) - 1
    numd = len(data_label)

    # create array to store all values before taking median
    yarr = np.arange(numy)
    xarr = np.arange(numx)
    zarr = np.arange(numz)
    darr = np.arange(numd)

    total = np.zeros((numd, numz, numy, numx))
    hits = np.zeros((numd, numz, numy, numx))

    if by_orbit:
        iterator = inst.orbits
    else:
        iterator = inst
    # do loop to iterate over given season
    for i, sat in enumerate(iterator):

        if len(sat.data) != 0:
            xind = np.digitize(sat.data[label1], binx) - 1
            for xi in xarr:
                xindex, = np.where(xind == xi)
                if len(xindex) > 0:
                    yData = sat.data.iloc[xindex]
                    yind = np.digitize(yData[label2], biny) - 1
                    for yj in yarr:
                        yindex, = np.where(yind == yj)
                        if len(yindex) > 0:
                            zData = yData.iloc[yindex]
                            zind = np.digitize(zData[label3], binz) - 1
                            for zk in zarr:
                                zindex, = np.where(zind == zk)
                                if len(zindex) > 0:
                                    for di in darr:
                                        indlab = zData.columns.get_loc(data_label[di])
                                        ddata = zData.iloc[zindex, indlab]
                                        idx, = np.where(np.isfinite(ddata))
                                        if len(idx) > 0:
                                            total[di, zk, yj, xi] += 1
                                            idx, = np.where(ddata > gate[di])
                                            if len(idx) > 0:
                                                hits[di, zk, yj, xi] += 1

    # all of the loading and storing data is done
    # prob = np.zeros((numz, numy, numx))
    prob = hits/total

    # make nicer dictionary output
    output = {}
    for i, label in enumerate(data_label):
        output[label] = {'prob': prob[i, :, :, :], 'count': total[i, :, :, :]}
        if returnBins:
            output[label]['bin_x'] = binx
            output[label]['bin_y'] = biny
            output[label]['bin_z'] = binz
    # clean up
    del iterator
    return output



"""
tests the pysat utils area
"""
import os
import tempfile
import warnings

from nose.tools import assert_raises, raises
import numpy as np
import pandas as pds

import pysat

import sys
if sys.version_info[0] >= 3:
    from importlib import reload as re_load
else:
    re_load = reload


# ----------------------------------
# test netCDF export file support

def prep_dir(inst=None):

    if inst is None:
        inst = pysat.Instrument(platform='pysat', name='testing')
    # create data directories
    try:
        os.makedirs(inst.files.data_path)
    except OSError:
        pass


def remove_files(inst):
    # remove any files
    temp_dir = inst.files.data_path
    for the_file in os.listdir(temp_dir):
        if (the_file == 'pysat_test_ncdf.nc'):
            file_path = os.path.join(temp_dir, the_file)
            if os.path.isfile(file_path):
                os.unlink(file_path)


def test_deprecation_warning_computational_form():
    """Test if computational form in utils is deprecated"""

    data = pds.Series([0, 1, 2])
    warnings.simplefilter("always")
    dslice1 = pysat.ssnl.computational_form(data)
    with warnings.catch_warnings(record=True) as war:
        dslice2 = pysat.utils.computational_form(data)

    assert (dslice1 == dslice2).all()
    assert len(war) >= 1
    assert war[0].category == DeprecationWarning


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.data_dir

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        pysat.utils.set_data_dir(self.data_path)

    #######################
    # test pysat data dir options
    def test_set_data_dir(self):
        """update data_dir"""
        pysat.utils.set_data_dir('.')
        check1 = (pysat.data_dir == '.')

        # Check if next load of pysat remembers the change
        pysat._files = re_load(pysat._files)
        pysat._instrument = re_load(pysat._instrument)
        re_load(pysat)
        check2 = (pysat.data_dir == '.')

        assert check1 & check2

    def test_set_data_dir_no_store(self):
        """update data_dir without storing"""
        pysat.utils.set_data_dir('.', store=False)
        check1 = (pysat.data_dir == '.')

        # Check if next load of pysat remembers old settings
        pysat._files = re_load(pysat._files)
        pysat._instrument = re_load(pysat._instrument)
        re_load(pysat)
        check2 = (pysat.data_dir == self.data_path)

        assert check1 & check2

    @raises(ValueError)
    def test_set_data_dir_wrong_path(self):
        """update data_dir with an invalid path"""
        pysat.utils.set_data_dir('not_a_directory', store=False)

    def test_initial_pysat_load(self):
        import shutil
        saved = False
        try:
            root = os.path.join(os.getenv('HOME'), '.pysat')
            new_root = os.path.join(os.getenv('HOME'), '.saved_pysat')
            shutil.move(root, new_root)
            saved = True
        except:
            pass

        re_load(pysat)

        try:
            if saved:
                # remove directory, trying to be careful
                os.remove(os.path.join(root, 'data_path.txt'))
                os.rmdir(root)
                shutil.move(new_root, root)
        except:
            pass

        assert True


class TestScaleUnits():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.deg_units = ["deg", "degree", "degrees", "rad", "radian",
                          "radians", "h", "hr", "hrs", "hours"]
        self.dist_units = ["m", "km", "cm"]
        self.vel_units = ["m/s", "cm/s", "km/s", 'm s$^{-1}$', 'cm s$^{-1}$',
                          'km s$^{-1}$', 'm s-1', 'cm s-1', 'km s-1']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.deg_units, self.dist_units, self.vel_units

    def test_scale_units_same(self):
        """ Test scale_units when both units are the same """

        scale = pysat.utils.scale_units("happy", "happy")

        assert scale == 1.0

    def test_scale_units_angles(self):
        """Test scale_units for angles """

        for out_unit in self.deg_units:
            scale = pysat.utils.scale_units(out_unit, "deg")

            if out_unit.find("deg") == 0:
                assert scale == 1.0
            elif out_unit.find("rad") == 0:
                assert scale == np.pi / 180.0
            else:
                assert scale == 1.0 / 15.0

    def test_scale_units_dist(self):
        """Test scale_units for distances """

        for out_unit in self.dist_units:
            scale = pysat.utils.scale_units(out_unit, "m")

            if out_unit == "m":
                assert scale == 1.0
            elif out_unit.find("km") == 0:
                assert scale == 0.001
            else:
                assert scale == 100.0

    def test_scale_units_vel(self):
        """Test scale_units for velocities """

        for out_unit in self.vel_units:
            scale = pysat.utils.scale_units(out_unit, "m/s")

            if out_unit.find("m") == 0:
                assert scale == 1.0
            elif out_unit.find("km") == 0:
                assert scale == 0.001
            else:
                assert scale == 100.0

    def test_scale_units_bad_output(self):
        """Test scale_units for unknown output unit"""

        assert_raises(ValueError, pysat.utils.scale_units, "happy", "m")
        try:
            pysat.utils.scale_units('happy', 'm')
        except ValueError as verr:
            assert str(verr).find('output unit') > 0

    def test_scale_units_bad_input(self):
        """Test scale_units for unknown input unit"""

        assert_raises(ValueError, pysat.utils.scale_units, "m", "happy")
        try:
            pysat.utils.scale_units('m', 'happy')
        except ValueError as verr:
            assert str(verr).find('input unit') > 0

    def test_scale_units_bad_match_pairs(self):
        """Test scale_units for mismatched input for all pairings"""

        assert_raises(ValueError, pysat.utils.scale_units, "m", "m/s")
        assert_raises(ValueError, pysat.utils.scale_units, "m", "deg")
        assert_raises(ValueError, pysat.utils.scale_units, "h", "km/s")

    def test_scale_units_bad_match_message(self):
        """Test scale_units error message for mismatched input"""

        assert_raises(ValueError, pysat.utils.scale_units, "m", "m/s")
        try:
            pysat.utils.scale_units('m', 'm/s')
        except ValueError as verr:
            assert str(verr).find('Cannot scale') >= 0
            assert str(verr).find('unknown units') < 0

    def test_scale_units_both_bad(self):
        """Test scale_units for bad input and output"""

        assert_raises(ValueError, pysat.utils.scale_units, "happy", "sad")
        try:
            pysat.utils.scale_units('happy', 'sad')
        except ValueError as verr:
            assert str(verr).find('unknown units') > 0


class TestBasicNetCDF4():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.data_dir

        # create temporary directory
        dir_name = tempfile.mkdtemp()
        pysat.utils.set_data_dir(dir_name, store=False)

        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing',
                                         sat_id='100',
                                         clean_level='clean')
        self.testInst.pandas_format = True

        # create testing directory
        prep_dir(self.testInst)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        remove_files(self.testInst)
        pysat.utils.set_data_dir(self.data_path, store=False)
        del self.testInst

    @raises(ValueError)
    def test_load_netcdf4_empty_filenames(self):
        pysat.utils.load_netcdf4(fnames=None)

    def test_basic_write_and_read_netcdf4_default_format(self):
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(2009, 1)
        self.testInst.to_netcdf4(outfile)

        loaded_inst, meta = \
            pysat.utils.load_netcdf4(outfile,
                                     pandas_format=self.testInst.pandas_format)
        self.testInst.data = \
            self.testInst.data.reindex(sorted(self.testInst.data.columns),
                                       axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns),
                                          axis=1)
        keys = self.testInst.data.columns

        for key in keys:
            assert(np.all(self.testInst[key] == loaded_inst[key]))

    def test_basic_write_and_read_netcdf4_mixed_case_format(self):
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(2009, 1)
        # modify data names in data
        original = sorted(self.testInst.data.columns)
        self.testInst.data = self.testInst.data.rename(str.upper, axis='columns')
        self.testInst.to_netcdf4(outfile, preserve_meta_case=True)

        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        self.testInst.data = \
            self.testInst.data.reindex(sorted(self.testInst.data.columns),
                                       axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)

        # check that names are lower case when written
        assert(np.all(original == loaded_inst.columns))

        for key in self.testInst.data.columns:
            assert(np.all(self.testInst[key] == loaded_inst[key.lower()]))

        # modify metadata names in data
        self.testInst.meta.data = self.testInst.meta.data.rename(str.upper, axis='index')
        # write file
        self.testInst.to_netcdf4(outfile, preserve_meta_case=True)
        # load file
        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)

        # check that names are upper case when written
        assert(np.all(sorted(self.testInst.data.columns) == sorted(loaded_inst.columns)))

    @raises(Exception)
    def test_write_netcdf4_duplicate_variable_names(self):
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(2009, 1)
        self.testInst['MLT'] = 1
        self.testInst.to_netcdf4(outfile, preserve_meta_case=True)

    def test_write_and_read_netcdf4_default_format_w_compression(self):
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(2009, 1)
        self.testInst.to_netcdf4(outfile, zlib=True)

        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        self.testInst.data = \
            self.testInst.data.reindex(sorted(self.testInst.data.columns),
                                       axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)

        for key in self.testInst.data.columns:
            assert (np.all(self.testInst[key] == loaded_inst[key]))

    def test_write_and_read_netcdf4_default_format_w_weird_epoch_name(self):
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(2009, 1)
        self.testInst.to_netcdf4(outfile, epoch_name='Santa')

        loaded_inst, meta = pysat.utils.load_netcdf4(outfile,
                                                     epoch_name='Santa')
        self.testInst.data = \
            self.testInst.data.reindex(sorted(self.testInst.data.columns),
                                       axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)

        for key in self.testInst.data.columns:
            assert (np.all(self.testInst[key] == loaded_inst[key]))

    def test_write_and_read_netcdf4_default_format_higher_order(self):
        # create a bunch of files by year and doy
        test_inst = pysat.Instrument('pysat', 'testing2d')
        prep_dir(test_inst)
        outfile = os.path.join(test_inst.files.data_path, 'pysat_test_ncdf.nc')
        test_inst.load(2009, 1)
        test_inst.to_netcdf4(outfile)
        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        test_inst.data = test_inst.data.reindex(sorted(test_inst.data.columns),
                                                axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)
        prep_dir(test_inst)

        # test Series of DataFrames
        test_list = []
        for frame1, frame2 in zip(test_inst.data['profiles'],
                                  loaded_inst['profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('profiles', inplace=True, axis=1)
        test_inst.data.drop('profiles', inplace=True, axis=1)

        # second series of frames
        for frame1, frame2 in zip(test_inst.data['alt_profiles'],
                                  loaded_inst['alt_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('alt_profiles', inplace=True, axis=1)
        test_inst.data.drop('alt_profiles', inplace=True, axis=1)

        # check series of series
        for frame1, frame2 in zip(test_inst.data['series_profiles'],
                                  loaded_inst['series_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))

        loaded_inst.drop('series_profiles', inplace=True, axis=1)
        test_inst.data.drop('series_profiles', inplace=True, axis=1)

        assert(np.all((test_inst.data == loaded_inst).all()))
        assert np.all(test_list)

    def test_write_and_read_netcdf4_default_format_higher_order_w_zlib(self):
        # create a bunch of files by year and doy
        test_inst = pysat.Instrument('pysat', 'testing2d')
        prep_dir(test_inst)
        outfile = os.path.join(test_inst.files.data_path, 'pysat_test_ncdf.nc')
        test_inst.load(2009, 1)
        test_inst.to_netcdf4(outfile, zlib=True)
        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        test_inst.data = test_inst.data.reindex(sorted(test_inst.data.columns),
                                                axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)
        prep_dir(test_inst)

        # test Series of DataFrames
        test_list = []
        for frame1, frame2 in zip(test_inst.data['profiles'],
                                  loaded_inst['profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('profiles', inplace=True, axis=1)
        test_inst.data.drop('profiles', inplace=True, axis=1)

        # second series of frames
        for frame1, frame2 in zip(test_inst.data['alt_profiles'],
                                  loaded_inst['alt_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('alt_profiles', inplace=True, axis=1)
        test_inst.data.drop('alt_profiles', inplace=True, axis=1)

        # check series of series
        for frame1, frame2 in zip(test_inst.data['series_profiles'],
                                  loaded_inst['series_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('series_profiles', inplace=True, axis=1)
        test_inst.data.drop('series_profiles', inplace=True, axis=1)

        assert (np.all((test_inst.data == loaded_inst).all()))
        assert np.all(test_list)

    def test_netcdf_prevent_attribute_override(self):
        """Test that attributes will not be overridden by default"""
        self.testInst.load(2009, 1)

        try:
            assert self.testInst.bespoke  # should raise
        except AttributeError:
            pass

        # instrument meta attributes immutable upon load
        assert not self.testInst.meta.mutable
        try:
            self.testInst.meta.bespoke = True
        except AttributeError:
            pass

    def test_netcdf_attribute_override(self):
        """Test that attributes in netcdf file may be overridden"""
        self.testInst.load(2009, 1)

        self.testInst.meta.mutable = True
        self.testInst.meta.bespoke = True

        self.testInst.meta.transfer_attributes_to_instrument(self.testInst)

        # ensure custom meta attribute assigned to instrument
        assert self.testInst.bespoke

        fname = 'output.nc'
        outfile = os.path.join(self.testInst.files.data_path, fname)
        self.testInst.to_netcdf4(outfile)

        data, meta = pysat.utils.load_netcdf4(outfile)

        # custom attribute correctly read from file
        assert meta.bespoke


class TestBasicNetCDF4xarray():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.data_dir

        # create temporary directory
        dir_name = tempfile.mkdtemp()
        pysat.utils.set_data_dir(dir_name, store=False)

        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing2d_xarray',
                                         sat_id='100',
                                         clean_level='clean')
        self.testInst.pandas_format = False

        # create testing directory
        prep_dir(self.testInst)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        remove_files(self.testInst)
        pysat.utils.set_data_dir(self.data_path, store=False)
        del self.testInst

    def test_basic_write_and_read_netcdf4_default_format(self):
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(2009, 1)
        self.testInst.data.attrs['new_attr'] = 1
        self.testInst.data.to_netcdf(outfile)

        loaded_inst, meta = \
            pysat.utils.load_netcdf4(outfile,
                                     pandas_format=self.testInst.pandas_format)
        keys = self.testInst.data.data_vars.keys()

        for key in keys:
            assert(np.all(self.testInst[key] == loaded_inst[key]))
        assert meta.new_attr == 1

    def test_load_netcdf4_pandas_3d_deprecation_warning(self):
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(2009, 1)
        self.testInst.data.attrs['new_attr'] = 1
        self.testInst.data.to_netcdf(outfile)

        warnings.simplefilter("always")
        with warnings.catch_warnings(record=True) as war:
            loaded_inst, meta = pysat.utils.load_netcdf4(outfile,
                                                         epoch_name='time',
                                                         pandas_format=True)
        assert len(war) >= 1
        assert war[0].category == DeprecationWarning



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------
""" Created as part of a Spring 2018 UTDesign project.
"""

import numpy as np


class Constellation(object):
    """Manage and analyze data from multiple pysat Instruments.

    Parameters
    ----------
    const_module : string
        Name of a pysat constellation module
    instruments : list-like
        A list of pysat Instruments to include in the Constellation

    Attributes
    ----------
    instruments : list
        A list of pysat Instruments that make up the Constellation
    bounds : (datetime/filename/None, datetime/filename/None)
        bounds for loading data, supply array_like for a season with gaps.
        Users may provide as a tuple or tuple of lists, but the attribute is
        stored as a tuple of lists for consistency

    """

    # -----------------------------------------------------------------------
    # Define the magic methods

    def __init__(self, const_module=None, instruments=None):
        """
        Constructs a Constellation given a list of instruments or the name of
        a file with a pre-defined constellation.

        Parameters
        ----------
        const_module : string
            Name of a pysat constellation module
        instruments : list-like
            A list of pysat Instruments to include in the Constellation

        Raises
        ------
        ValueError
            When `instruments` is not list-like

        Note
        ----
        Omit `instruments` and `const_module` to create an empty constellation.

        """

        # Include Instruments from the constellation module, if it exists
        if const_module is not None:
            if not hasattr(const_module, 'instruments'):
                raise AttributeError("missing required attribute 'instruments'")
            self.instruments = const_module.instruments
        else:
            self.instruments = []

        # Add any Instruments provided in the list
        if instruments is not None:
            test_instruments = np.asarray(instruments)
            if test_instruments.shape == ():
                raise ValueError('instruments argument must be list-like')

            self.instruments.extend(list(instruments))

        return

    def __getitem__(self, *args, **kwargs):
        """
        Look up a member Instrument by index.

        """

        return self.instruments.__getitem__(*args, **kwargs)

    def __repr__(self):
        """ Print the basic Constellation properties

        """

        out_str = "".join(["pysat.Constellation(instruments=",
                           "{:})".format(self.instruments)])
        return out_str

    def __str__(self):
        """ Print names of instruments within constellation

        """

        output_str = 'pysat Constellation object:\n'
        output_str += '---------------------------\n'

        ninst = len(self.instruments)

        if ninst > 0:
            output_str += "\nIndex Platform Name Tag Inst_ID\n"
            output_str += "-------------------------------\n"
            for i, inst in enumerate(self.instruments):
                output_str += "{:d} '{:s}' '{:s}' '{:s}' '{:s}'\n".format(
                    i, inst.platform, inst.name, inst.tag, inst.inst_id)
        else:
            output_str += "No loaded Instruments\n"

        return output_str

    # -----------------------------------------------------------------------
    # Define the public methods and properties

    @property
    def bounds(self):
        return self.instruments[0].bounds

    @bounds.setter
    def bounds(self, value=None):
        """ Sets boundaries for all Instruments in Constellation

        Parameters
        ----------
        value : tuple or NoneType
            Tuple containing starting time and ending time for Instrument
            bounds attribute or None (default=None)

        """

        for instrument in self.instruments:
            instrument.bounds = value

        return

    def custom_attach(self, *args, **kwargs):
        """Register a function to modify data of member Instruments.

        Parameters
        ----------
        *args : list reference
            References a list of input arguments
        **kwargs : dict reference
            References a dict of input keyword arguments

        See Also
        ---------
        Instrument.custom_attach : base method for attaching custom functions

        """

        for instrument in self.instruments:
            instrument.custom_attach(*args, **kwargs)

        return

    def custom_clear(self):
        """Clear the custom function list

        See Also
        ---------
        Instrument.custom_clear : base method for clearing custom functions

        """

        for instrument in self.instruments:
            instrument.custom_clear()

        return

    def load(self, *args, **kwargs):
        """ Load instrument data into Instrument object.data

        Parameters
        ----------
        *args : list reference
            References a list of input arguments
        **kwargs : dict reference
            References a dict of input keyword arguments

        See Also
        ---------
        Instrument.load : base method for loading Instrument data

        """

        for instrument in self.instruments:
            instrument.load(*args, **kwargs)

        return



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------

import copy
import datetime as dt
from functools import partial
import numpy as np
import os
import weakref

import pandas as pds

import pysat  # Needed to access pysat.params across reimports
from pysat.utils import files as futils
from pysat.utils.time import filter_datetime_input
from pysat.instruments.methods import general

logger = pysat.logger


class Files(object):
    """Maintains collection of files and associated methods.

    Parameters
    ----------
    inst : pysat.Instrument
        Instrument object
    directory_format : str or NoneType
        Directory naming structure in string format. Variables such as
        platform, name, tag, and inst_id will be filled in as needed using
        python string formatting. The default directory structure would be
        expressed as '{platform}/{name}/{tag}/{inst_id}'. If None, the default
        directory structure is used (default=None)
    update_files : boolean
        If True, immediately query filesystem for instrument files and
        store (default=False)
    file_format : str or NoneType
        File naming structure in string format.  Variables such as year,
        month, day, and inst_id will be filled in as needed using python string
        formatting.  The default file format structure is supplied in the
        instrument list_files routine. (default=None)
    write_to_disk : boolean
        If true, the list of Instrument files will be written to disk.
        (default=True)
    ignore_empty_files : boolean
        If True, the list of files found will be checked to ensure the
        filesizes are greater than zero. Empty files are removed from the
        stored list of files. (default=False)

    Attributes
    ----------
    home_path : str
        Path to the pysat information directory.
    data_path : str
        Path to the top-level directory containing instrument files,
        selected from data_paths.
    data_paths: list of str
        Available paths that pysat will use when looking for files. The
        class uses the first directory with relevant data, stored in data_path.
    files : pds.Series
        Series of data files, indexed by file start time
    inst_info : dict
        Contains pysat.Instrument parameters 'platform', 'name', 'tag',
        and 'inst_id', identifying the source of the files.
    list_files_creator : functools.partial or NoneType
        Experimental feature for Instruments that internally generate data
        and thus don't have a defined supported date range.
    list_files_rtn : method
        Method used to locate relevant files on the local system. Provided
        by associated pysat.Instrument object.
    multi_file_day : boolean
        Flag copied from associated pysat.Instrument object that indicates
        when data for day n may be found in files for days n-1, or n+1
    start_date : datetime or NoneType
        Date of first file, used as default start bound for instrument
        object, or None if no files are loaded.
    stop_date : datetime or NoneType
        Date of last file, used as default stop bound for instrument
        object, or None if no files are loaded.
    stored_file_name : str
        Name of the hidden file containing the list of archived data files
        for this instrument.
    sub_dir_path : str
        `directory_format` string formatted for the local system.


    Note
    ----
    Interfaces with the `list_files` method for a given instrument
    support module to create an ordered collection of files in time,
    used primarily by the pysat.Instrument object to identify files
    to be loaded. The Files class mediates access to the files by
    datetime and contains helper methods for determining the presence of
    new files and filtering out empty files.

    User should generally use the interface provided by a pysat.Instrument
    instance. Exceptions are the classmethod from_os, provided to assist
    in generating the appropriate output for an instrument routine.

    Examples
    --------
    ::

        # convenient file access
        inst = pysat.Instrument(platform=platform, name=name, tag=tag,
                                inst_id=inst_id)
        # first file
        inst.files[0]

        # files from start up to stop (exclusive on stop)
        start = dt.datetime(2009,1,1)
        stop = dt.datetime(2009,1,3)
        print(inst.files[start:stop])

        # files for date
        print(inst.files[start])

        # files by slicing
        print(inst.files[0:4])

        # get a list of new files
        # new files are those that weren't present the last time
        # a given instrument's file list was stored
        new_files = inst.files.get_new()

        # search pysat appropriate directory for instrument files and
        # update Files instance.
        inst.files.refresh()

    """

    # -----------------------------------------------------------------------
    # Define the magic methods

    def __init__(self, inst, directory_format=None, update_files=False,
                 file_format=None, write_to_disk=True,
                 ignore_empty_files=False):

        # Set the hidden variables
        self.update_files = update_files

        # Location of directory to store file information in
        self.home_path = os.path.join(pysat.pysat_dir, 'instruments')

        # Assign base default dates and an empty list of files
        self.start_date = None
        self.stop_date = None
        self.files = pds.Series(None, dtype='object')

        # Grab Instrument info
        self.inst_info = {'platform': inst.platform, 'name': inst.name,
                          'tag': inst.tag, 'inst_id': inst.inst_id,
                          'inst_module': inst.inst_module,
                          'inst': weakref.proxy(inst)}

        self.multi_file_day = inst.multi_file_day

        # Begin with presumption that the list_files_rtn is a typical
        # function that returns a Series of filenames. Some generated
        # data sets employ a function that creates filenames on-the-fly.
        self.list_files_creator = None

        # Set the location of stored files
        self.stored_file_name = '_'.join((self.inst_info['platform'],
                                          self.inst_info['name'],
                                          self.inst_info['tag'],
                                          self.inst_info['inst_id'],
                                          'stored_file_info.txt'))

        # Set the path for sub-directories under pysat data path
        if directory_format is None:
            # Assign stored template if user doesn't provide one.
            directory_format = pysat.params['directory_format']
        self.directory_format = directory_format

        # Set the user-specified file format
        self.file_format = file_format

        # Construct the subdirectory path
        self.sub_dir_path = os.path.normpath(
            self.directory_format.format(**self.inst_info))

        # Ensure we have at least one path for pysat data directory
        if len(pysat.params['data_dirs']) == 0:
            raise NameError(" ".join(("pysat's `data_dirs` hasn't been set.",
                                      "Please set a top-level directory",
                                      "path to store data using",
                                      "`pysat.params['data_dirs'] = path`")))

        # Get list of potential data directory paths from pysat. Construct
        # possible locations for data. Ensure path always ends with directory
        # separator.
        self.data_paths = [os.path.join(pdir, self.sub_dir_path)
                           for pdir in pysat.params['data_dirs']]
        self.data_paths = [os.path.join(os.path.normpath(pdir), '')
                           for pdir in self.data_paths]

        # Only one of the above paths will actually be used when loading data.
        # The actual value of data_path is determined in refresh().
        # If there are files present, then that path is stored along with a
        # list of found files in ~/.pysat. This stored info is retrieved by
        # _load. We start here with the first directory for cases where there
        # are no files.
        self.data_path = self.data_paths[0]

        # Set the preference of writing the file list to disk or not
        self.write_to_disk = write_to_disk
        if not self.write_to_disk:
            # Use blank memory rather than loading from disk
            self._previous_file_list = pds.Series([], dtype='a')
            self._current_file_list = pds.Series([], dtype='a')

        # Set the preference to ignore or include empty files
        self.ignore_empty_files = ignore_empty_files

        if self.inst_info['platform'] != '':
            # Only load filenames if this is associated with a real
            # pysat.Instrument instance, not pysat.Instrument().
            if self.update_files:
                # Refresh filenames as directed by user
                self.refresh()
            else:
                # Load stored file info
                file_info = self._load()
                if file_info.empty:
                    # Didn't find stored information. Search local system.
                    # If list_files_rtn returns a dict to create
                    # filenames as needed that is handled in refresh.
                    self.refresh()
                else:
                    # Attach the data loaded
                    self._attach_files(file_info)
        return

    def __repr__(self):
        """ Representation of the class and its current state
        """
        inst_repr = self.inst_info['inst'].__repr__()

        out_str = "".join(["pysat.Files(", inst_repr, ", directory_format=",
                           "'{:}'".format(self.directory_format),
                           ", update_files=",
                           "{:}, file_format=".format(self.update_files),
                           "{:}, ".format(self.file_format.__repr__()),
                           "write_to_disk={:}, ".format(self.write_to_disk),
                           "ignore_empty_files=",
                           "{:})".format(self.ignore_empty_files)])

        return out_str

    def __str__(self):
        """ Description of the class and its contents
        """

        num_files = len(self.files)
        output_str = 'Local File Statistics\n'
        output_str += '---------------------\n'
        output_str += 'Number of files: {:d}\n'.format(num_files)

        if num_files > 0:
            output_str += 'Date Range: '
            output_str += self.files.index[0].strftime('%d %B %Y')
            output_str += ' --- '
            output_str += self.files.index[-1].strftime('%d %B %Y')

        return output_str

    def __eq__(self, other):
        """Perform equality check

        Parameters
        ----------
        other : any
            Other object to compare for equality

        Returns
        -------
        bool
            True if objects are identical, False if they are not

        """
        # Check if the other object has the same type
        if not isinstance(other, self.__class__):
            return False

        # If the type is the same then check everything that is attached to
        # the Files object. Includes attributes, methods, variables, etc.
        checks = []
        key_check = []
        for key in self.__dict__.keys():
            key_check.append(key)
            # Confirm each object has the same keys
            if key in other.__dict__.keys():
                # Define default comparison.
                if key not in ['files', '_previous_file_list',
                               '_current_file_list', 'inst_info']:
                    test = np.all(self.__dict__[key] == other.__dict__[key])
                    checks.append(test)

                else:
                    if key not in ['inst_info']:
                        # Comparing one of the stored pandas Series
                        try:
                            # Comparison only works for identically-labeled
                            # series.
                            check = np.all(self.__dict__[key]
                                           == other.__dict__[key])
                            checks.append(check)
                        except ValueError:
                            # If there is an error they aren't the same.
                            return False

                    elif key == 'inst_info':
                        ichecks = []
                        for ii_key in self.inst_info.keys():
                            if ii_key != 'inst':
                                # Standard attribute check
                                ichecks.append(self.inst_info[ii_key]
                                               == other.inst_info[ii_key])

                            else:
                                # Don't want a recursive check on 'inst', which
                                # contains Files. If the string representations
                                # are the same we consider them the same.
                                try:
                                    oinst = other.inst_info[ii_key]
                                    ichecks.append(str(self.inst_info[ii_key])
                                                   == str(oinst))
                                except AttributeError:
                                    # If one object is missing a required key
                                    return False
                        checks.append(np.all(ichecks))

            else:
                # other did not have an key that self did
                return False

        # Confirm that Files object `other` doesn't have extra terms
        for key in other.__dict__.keys():
            if key not in self.__dict__.keys():
                return False

        test_data = np.all(checks)

        return test_data

    def __getitem__(self, key):
        """ Retrieve items from the files attribute

        Parameters
        ----------
        key : int, list, slice, dt.datetime
            Key for locating files from a pandas Series indexed by time

        Returns
        -------
        out : pds.Series
           Subset of the files as a Series

        Raises
        ------
        IndexError
            If data is outside of file bounds

        Note
        ----
        Slicing via date and index filename is inclusive slicing, date and
        index are normal non-inclusive end point

        """
        if self.list_files_creator is not None:
            # Return filename generated on demand
            out = self.list_files_creator(key)

        elif isinstance(key, slice):
            try:
                try:
                    # Assume key is integer (including list or slice)
                    out = self.files.iloc[key]
                except TypeError:
                    # The key must be something else, use alternative access
                    out = self.files.loc[key]
            except IndexError as err:
                raise IndexError(''.join((str(err), '\n',
                                          'Date requested outside file ',
                                          'bounds.')))

            if isinstance(key.start, dt.datetime):
                # Enforce exclusive slicing on datetime
                if len(out) > 1:
                    if out.index[-1] >= key.stop:
                        out = out[:-1]
                elif len(out) == 1:
                    if out.index[0] >= key.stop:
                        out = pds.Series([], dtype='a')
        else:
            try:
                # Assume key is integer (including list or slice)
                out = self.files.iloc[key]
            except TypeError:
                # The key must be something else, use alternative access
                out = self.files.loc[key]

        return out

    # -----------------------------------------------------------------------
    # Define the hidden methods

    def _filter_empty_files(self, path):
        """Update the file list (self.files) with empty files removed

        Parameters
        ----------
        path : str
            Path to top-level containing files

        """

        keep_index = []
        for i, fname in enumerate(self.files):
            # Create full path for each file
            full_fname = os.path.join(path, fname)

            # Ensure the file exists
            if os.path.isfile(full_fname):
                # Check for size
                if os.path.getsize(full_fname) > 0:
                    # Store if not empty
                    keep_index.append(i)

        # Remove filenames as needed
        dropped_num = len(self.files.index) - len(keep_index)
        if dropped_num > 0:
            logger.warning(' '.join(('Removing {:d}'.format(dropped_num),
                                     'empty files from Instrument list.')))
            self.files = self.files.iloc[keep_index]

        return

    def _attach_files(self, files_info):
        """Attaches stored file lists to self.files

        Parameters
        ---------
        files_info : pds.Series
            Stored file information, filenames indexed by datetime

        Note
        ----
        Updates the file list (files), start_date, and stop_date attributes
        of the Files class object.

        """

        if not files_info.empty:
            # Attach data
            self.files = files_info

            # Ensure times are unique.
            self._ensure_unique_file_datetimes()

            # Filter for empty files.
            if self.ignore_empty_files:
                self._filter_empty_files(path=self.data_path)

            # Extract date information from first and last files
            if not self.files.empty:
                self.start_date = filter_datetime_input(self.files.index[0])
                self.stop_date = filter_datetime_input(self.files.index[-1])
            else:
                # No files found
                self.start_date = None
                self.stop_date = None
        else:
            # No files found
            self.start_date = None
            self.stop_date = None

            # Convert to object type if Series is empty.  This allows for
            # `==` equality checks with strings
            self.files = files_info.astype(np.dtype('O'))

        return

    def _ensure_unique_file_datetimes(self):
        """Update the file list (self.files) to ensure uniqueness"""

        # Check if files are unique.
        unique_files = len(self.files.index.unique()) == len(self.files)

        if not self.multi_file_day and not unique_files:
            # Give user feedback about the issue
            estr = ''.join(['Duplicate datetimes in stored filename ',
                            'information.\nKeeping one of each ',
                            'of the duplicates, dropping the rest. ',
                            'Please ensure the file datetimes ',
                            'are unique at the microsecond level.'])
            logger.warning(estr)
            ind = self.files.index.duplicated()
            logger.warning(self.files.index[ind].unique())

            # Downselect to unique file datetimes
            idx = np.unique(self.files.index, return_index=True)
            self.files = self.files.iloc[idx[1]]

        return

    def _store(self):
        """Store currently loaded filelist for instrument onto filesystem
        """

        stored_name = self.stored_file_name

        # Check if current file data is different than stored file list. If so,
        # move file list to previous file list, store current to file. If not,
        # do nothing
        stored_files = self._load(update_path=False)
        if len(stored_files) != len(self.files):
            # The number of items is different, things are new
            new_flag = True
        else:
            # The number of items is the same, check specifically for equality
            if stored_files.eq(self.files).all():
                new_flag = False
            else:
                # Stored and new data are not equal, there are new files
                new_flag = True

        if new_flag:
            if self.write_to_disk:
                # Save the previous data in a backup file
                prev_name = os.path.join(self.home_path, 'archive', stored_name)
                stored_files.to_csv(prev_name,
                                    date_format='%Y-%m-%d %H:%M:%S.%f',
                                    header=[self.data_path])

                # Overwrite the old reference file with the new file info
                self.files.to_csv(os.path.join(self.home_path, stored_name),
                                  date_format='%Y-%m-%d %H:%M:%S.%f',
                                  header=[self.data_path])
            else:
                # Update the hidden File attributes
                self._previous_file_list = stored_files
                self._current_file_list = self.files.copy()

        return

    def _load(self, prev_version=False, update_path=True):
        """Load stored filelist

        Parameters
        ----------
        prev_version : boolean
            if True, will load previous version of file list
        update_path : boolean
            If True, the path written to stored info will be
            assigned to self.data_path. (default=True)

        Returns
        -------
        pandas.Series
            File path names, indexed by datetime. Series is empty if no
            files are found.

        """

        fname = self.stored_file_name
        if prev_version:
            # Archived file list storage filename
            fname = os.path.join(self.home_path, 'archive', fname)
        else:
            # Current file list storage filename
            fname = os.path.join(self.home_path, fname)

        if os.path.isfile(fname) and (os.path.getsize(fname) > 0):
            if self.write_to_disk:
                # Load data stored on the local drive.
                loaded = pds.read_csv(fname, index_col=0, parse_dates=True,
                                      squeeze=True, header=0)
                if update_path:
                    # Store the data_path from the .csv onto Files
                    self.data_path = loaded.name

                # Ensure the name of returned Series is None for consistency
                loaded.name = None

                return loaded
            else:
                # Grab content from memory rather than local disk.
                if prev_version:
                    return self._previous_file_list
                else:
                    return self._current_file_list
        else:
            # Storage file not present.
            return pds.Series([], dtype='a')

    def _remove_data_dir_path(self, file_series=None):
        """Remove the data directory path from filenames

        Parameters
        ----------
        file_series : pds.Series or NoneType
            Series of filenames (potentially with file paths)
            (default=None)

        Returns
        -------
        pds.series or None
            If `file_series` is a Series, removes the data path from the
            filename, if present.  Returns None if `path_input` is None.

        """
        out = None
        if file_series is not None:
            # Ensure there is a directory divider at the end of the path
            split_str = os.path.join(self.data_path, '')

            # Remove the data path from all filenames in the Series
            out = file_series.apply(lambda x: x.split(split_str)[-1])

        return out

    # -----------------------------------------------------------------------
    # Define the public methods and properties

    def copy(self):
        """Provide a deep copy of object

        Returns
        -------
        Files class instance
            Copy of self

        """
        # The copy module does not copy modules. Treat self.inst_info
        # differently since it possibly contains a python module, plus
        # it also contains a weakref back to Instrument.  Because the Instrument
        # reference contains another Files object, it could cause the creation
        # of an infinite, recursive copy.
        saved_info = self.inst_info
        self.inst_info = None

        # Copy everything but the problematic info
        files_copy = copy.deepcopy(self)

        # Restore the saved information, then copy over items that can be copied
        self.inst_info = saved_info
        files_copy.inst_info = {}
        for key in saved_info.keys():
            if key not in ['inst', 'inst_module']:
                files_copy.inst_info[key] = copy.deepcopy(self.inst_info[key])

        # Can't copy the weakreference
        files_copy.inst_info['inst'] = self.inst_info['inst']

        # Can't copy the module
        files_copy.inst_info['inst_module'] = self.inst_info['inst_module']
        return files_copy

    def refresh(self):
        """Update list of files, if there are changes.

        Note
        ----
        Calls underlying list_files_rtn for the particular science instrument.
        Typically, these routines search in the pysat provided path,
        pysat_data_dir/platform/name/tag/inst_id, where pysat_data_dir is set by
        pysat.utils.set_data_dir(path=path).

        """

        # Let interested users know pysat is searching for
        info_str = '{platform} {name} {tag} {inst_id}'.format(
            **self.inst_info)
        info_str = " ".join(("pysat is searching for", info_str, "files."))
        info_str = " ".join(info_str.split())  # Remove duplicate whitespace
        logger.info(info_str)

        # Check all potential directory locations for files.
        # Stop as soon as we find some.
        for path in self.data_paths:
            list_files_rtn = self.inst_info['inst']._list_files_rtn
            kwarg_inputs = self.inst_info['inst'].kwargs['list_files']
            new_files = list_files_rtn(tag=self.inst_info['tag'],
                                       inst_id=self.inst_info['inst_id'],
                                       data_path=path,
                                       format_str=self.file_format,
                                       **kwarg_inputs)

            # Check if list_files_rtn is actually returning filename or a
            # dict to be passed to filename creator function.
            if isinstance(new_files, dict):
                self.list_files_creator = partial(general.filename_creator,
                                                  **new_files)

                # Instrument iteration methods require a date range.
                self.start_date = filter_datetime_input(new_files['start_date'])
                self.stop_date = filter_datetime_input(new_files['stop_date'])

                # To really support iteration, we may need to create a generator
                # function that'll create a fake list of files as needed.
                # It would have to function in place of self.files. Is
                # there truly a point to this?
                return

            # Ensure the name of returned Series is None for consistency
            new_files.name = None

            # If we find some files, this is the one directory we store.
            # If I don't remove the directory paths then loading by filename
            # becomes more of a challenge. Plus, more memory to store, more
            # difficult for a human to parse when browsing a list, etc. The
            # approach here provides for most of the potential functionality
            # of multiple directories while still leaving the 'single' directory
            # focus and features of the original pysat intact.
            if not new_files.empty:
                self.data_path = path
                new_files = self._remove_data_dir_path(new_files)
                break

        # Feedback to info on number of files located
        logger.info('Found {:d} local files.'.format(len(new_files)))

        if not new_files.empty:
            # Sort files to ensure they are in order
            new_files = new_files.sort_index()
        elif pysat.params['warn_empty_file_list']:
            # Warn user if no files found, if pysat.param set
            pstrs = "\n".join(self.data_paths)
            estr = "".join(("Unable to find any files that match the supplied ",
                            "template: ", self.file_format, "\n",
                            "In the following directories: \n", pstrs))
            logger.warning(estr)

        # Attach Series of files to the class object
        self._attach_files(new_files)

        # Store to disk, if enabled for this class
        self._store()
        return

    def set_top_level_directory(self, path):
        """Sets top-level data directory.

        Sets a valid self.data_path using provided top-level directory
        path and the associated pysat subdirectories derived from the
        directory_format attribute as stored in self.sub_dir_path.

        Parameters
        ----------
        path : str
            Top-level path to use when looking for files. Must be in
            pysat.params['data_dirs']

        Note
        ----
        If there are Instrument files on the system under a top-level
        directory other than `path`, then, under certain conditions,
        self.data_path may be later updated by the object to point back
        to the directory with files.

        """

        if path not in pysat.params['data_dirs']:
            estr = "Supplied path not in `pysat.params['data_dirs']`"
            raise ValueError(estr)
        else:
            self.data_path = os.path.join(path, self.sub_dir_path)

        return

    def get_new(self):
        """List new files since last recorded file state.

        Returns
        -------
        pandas.Series
           A datetime-index Series of all new fileanmes since the last known
           change to the files.

        Note
        ----
        pysat stores filenames in the user_home/.pysat directory. Filenames are
        stored if there is a change and either update_files is True at
        instrument object level or files.refresh() is called.

        """

        # Refresh file series
        self.refresh()

        # Load current and previous set of files
        new_file_series = self._load(update_path=False)
        old_file_series = self._load(prev_version=True, update_path=False)

        # Select files that are in the new series and not the old series
        new_files = new_file_series[-new_file_series.isin(old_file_series)]

        return new_files

    def get_index(self, fname):
        """Return index for a given filename.

        Parameters
        ----------
        fname : string
            Filename for the desired time index

        Note
        ----
        If fname not found in the file information already attached
        to the instrument.files instance, then a files.refresh() call
        is made.

        """

        idx, = np.where(fname == self.files)
        if len(idx) == 0:
            # Filename not in index, try reloading files from disk
            self.refresh()
            idx, = np.where(fname == np.array(self.files))

            if len(idx) == 0:
                raise ValueError(' '.join(('Could not find "{:}"'.format(fname),
                                           'in available file list. Valid',
                                           'Example:', self.files.iloc[0])))

        # Return a scalar rather than array - otherwise introduces array to
        # index warnings.
        return idx[0]

    def get_file_array(self, start, stop):
        """Return a list of filenames between and including start and stop.

        Parameters
        ----------
        start: array_like or single string
            filenames for start of returned filelist
        stop: array_like or single string
            filenames inclusive of the ending of list provided by the stop time

        Returns
        -------
        files : list
            A list of filenames between and including start and stop times
            over all intervals.

        Note
        ----
        `start` and `stop` must be of the same type: both array-like or both
        strings

        """

        starts = np.asarray(start)
        if starts.shape == ():
            starts = [starts.tolist()]
        elif starts.shape[0] > 1:
            starts = starts.squeeze().tolist()
        elif starts.shape[0] == 1:
            starts = starts.tolist()

        stops = np.asarray(stop)
        if stops.shape == ():
            stops = [stops.tolist()]
        elif stops.shape[0] > 1:
            stops = stops.squeeze().tolist()
        elif stops.shape[0] == 1:
            stops = stops.tolist()

        # Selection is treated differently if start/stop are iterable or not
        files = []
        for (sta, stp) in zip(starts, stops):
            id1 = self.get_index(sta)
            id2 = self.get_index(stp)
            files.extend(self.files.iloc[id1:(id2 + 1)])

        return files

    @classmethod
    def from_os(cls, data_path=None, format_str=None,
                two_digit_year_break=None, delimiter=None):
        """
        Produces a list of files and and formats it for Files class.

        Parameters
        ----------
        data_path : string
            Top level directory to search files for. This directory
            is provided by pysat to the instrument_module.list_files
            functions as data_path.
        format_str : string with python format codes
            Provides the naming pattern of the instrument files and the
            locations of date information so an ordered list may be produced.
            Supports 'year', 'month', 'day', 'hour', 'minute', 'second',
            'version', 'revision', and 'cycle'
            Ex: 'cnofs_cindi_ivm_500ms_{year:4d}{month:02d}{day:02d}_v01.cdf'
        two_digit_year_break : int or None
            If filenames only store two digits for the year, then
            '1900' will be added for years >= two_digit_year_break
            and '2000' will be added for years < two_digit_year_break.
            If None, then four-digit years are assumed. (default=None)
        delimiter : string or NoneType
            Delimiter string upon which files will be split (e.g., '.'). If
            None, filenames will be parsed presuming a fixed width format.
            (default=None)

        Returns
        -------
        pds.Series
            A Series of filenames indexed by time. See
            `pysat.utils.files.process_parsed_filenames` for details.

        Note
        ----
        Requires fixed_width or delimited filename

        Does not produce a Files instance, but the proper output from
        instrument_module.list_files method.

        The '?' may be used to indicate a set number of spaces for a variable
        part of the name that need not be extracted.
        'cnofs_cindi_ivm_500ms_{year:4d}{month:02d}{day:02d}_v??.cdf'

        """

        if data_path is None:
            raise ValueError(" ".join(("Must supply instrument directory path",
                                       "(dir_path)")))

        # Parse format string to figure out which search string should be used
        # to identify files in the filesystem. Different option required if
        # filename is delimited
        wildcard = False if delimiter is None else True
        search_dict = futils.construct_searchstring_from_format(
            format_str, wildcard=wildcard)
        search_str = search_dict['search_string']

        # Perform the local file search
        files = futils.search_local_system_formatted_filename(data_path,
                                                              search_str)

        # Use the file list to extract the information. Pull data from the
        # areas identified by format_str
        if delimiter is None:
            stored = futils.parse_fixed_width_filenames(files, format_str)
        else:
            stored = futils.parse_delimited_filenames(files, format_str,
                                                      delimiter)

        # Process the parsed filenames and return a properly formatted Series
        return futils.process_parsed_filenames(stored, two_digit_year_break)



# -*- coding: utf-8 -*-.
"""Supports generalized access to Madrigal Data.

To use this routine, you need to know both the Madrigal Instrument code
as well as the data tag numbers that Madrigal uses to uniquely identify
data sets. Using these codes, the methods.madrigal.py routines will
be used to support downloading and loading of data.

Downloads data from the Madrigal Database.

Warnings
--------
    All data downloaded under this general support is placed in the same
    directory, pysat_data_dir/madrigal/pandas/. For technical reasons,
    the file search algorithm for pysat's Madrigal support is set to permissive
    defaults. Thus, all instrument files downloaded via this interface will be
    picked up by the madrigal pandas pysat Instrument object unless the
    file_format keyword is used at instantiation.

    Files can be safely downloaded without knowing the file_format keyword,
    or equivalently, how Madrigal names the files. See `Examples` for more.

Parameters
----------
platform : string
    'madrigal'
name : string
    'pandas'
tag : string
    ''

Examples
--------
::
    # for isolated use of a madrigal data set
    import pysat
    # download DMSP data from Madrigal
    dmsp = pysat.Instrument('madrigal', 'pandas',
                            madrigal_code=8100,
                            madrigal_tag=10241)
    dmsp.download(pysat.datetime(2017, 12, 30), pysat.datetime(2017, 12, 31),
                  user='Firstname+Lastname', password='email@address.com')
    dmsp.load(2017,363)

    # for users that plan on using multiple Madrigal datasets
    # using this general interface then an additional parameter
    # should be supplied upon instrument instantiation (file_format)

    # pysat needs information on how to parse filenames from Madrigal
    # for the particular instrument under study.
    # When starting from scratch (no files), this is a two step process.
    # First, get atleast one file from Madrigal, using the steps above
    # using the file downloaded. Using the filename, convert it to a template
    # string
    # and pass that to pysat when instantiating future Instruments.

    # For example, one of the files downloaded above is
    # dms_ut_19980101_11.002.hdf5
    # pysat needs a template for how to pull out the year, month, day, and,
    # if available, hour, minute, second, etc.
    # the format/template string for this instrument is
    # 'dms_ut_{year:4d}{month:02d}{day:02d}_12.002.hdf5', following
    # python standards for string templates/Formatters
    # https://docs.python.org/2/library/string.html

    # the complete instantiation for this instrument is
    dmsp = pysat.Instrument('madrigal', 'pandas',
                            madrigal_code=8100,
                            madrigal_tag=10241,
                            file_format='dms_ut_{year:4d}{month:02d}{day:02d}_11.002.hdf5')

Note
----
    Please provide name and email when downloading data with this routine.

"""

from __future__ import print_function
from __future__ import absolute_import

import functools
import pysat
from pysat.instruments.methods import madrigal as mad_meth
from pysat.instruments.methods import nasa_cdaweb as cdw

import logging
logger = logging.getLogger(__name__)

platform = 'madrigal'
name = 'pandas'
tags = {'': 'General Madrigal data access loaded into pysat via pandas.'}
sat_ids = {'': list(tags.keys())}
# need to sort out test day setting for unit testing
_test_dates = {'': {'': pysat.datetime(2010, 1, 19)}}

# support list files routine
# use the default CDAWeb method
#########
# need a way to get the filename strings for a particular instrument
# I've put in wildcards for now....
#########
jro_fname1 = '*{year:4d}{month:02d}{day:02d}'
jro_fname2 = '.{version:03d}.hdf5'
supported_tags = {ss: {'': jro_fname1 + "*" + jro_fname2}
                  for ss in sat_ids.keys()}
list_files = functools.partial(cdw.list_files,
                               supported_tags=supported_tags)

# let pysat know that data is spread across more than one file
# multi_file_day=True

# Set to False to specify using xarray (not using pandas)
# Set to True if data will be returned via a pandas DataFrame
pandas_format = True

# support load routine
load = mad_meth.load

# support download routine
# real download attached during init
# however, pysat requires a method before we get there
download = mad_meth.download


def init(self):
    """Initializes the Instrument object in support of Madrigal access

    Runs once upon instantiation.

    Parameters
    ----------
    self : pysat.Instrument
        This object

    """

    logger.info(mad_meth.cedar_rules())

    code = self.kwargs['madrigal_code']
    tag = self.kwargs['madrigal_tag']
    self._download_rtn = functools.partial(_general_download,
                                           inst_code=str(code),
                                           kindat=str(tag))
    return


def _general_download(date_array, tag='', sat_id='', data_path=None, user=None,
                      password=None, inst_code=None, kindat=None):
    """Downloads data from Madrigal.

    Method will be partially set using functools.partial. Intended to
    have the same call structure as normal instrument download routine.
    Upon Instrument instantiation this routine will be set to
    parameters specific to a Madrigal data set. It will then work like
    a standard download call.

    Parameters
    ----------
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need not
        be contiguous.
    tag : string
        Tag identifier used for particular dataset. This input is provided by
        pysat. (default='')
    sat_id : string
        Satellite ID string identifier used for particular dataset. This input
        is provided by pysat. (default='')
    data_path : string
        Path to directory to download data to. (default=None)
    user : string
        User string input used for download. Provided by user and passed via
        pysat. If an account is required for dowloads this routine here must
        error if user not supplied. (default=None)
    password : string
        Password for data download. (default=None)
    inst_code : int
        Madrigal integer code used to identify platform (default=None)
    kindat : int
        Madrigal integer code used to identify data set (default=None)

    Note
    ----
    The user's names should be provided in field user. Ruby Payne-Scott should
    be entered as Ruby+Payne-Scott

    The password field should be the user's email address. These parameters
    are passed to Madrigal when downloading.

    The affiliation field is set to pysat to enable tracking of pysat
    downloads.

    """
    mad_meth.download(date_array, inst_code=inst_code, kindat=kindat,
                      data_path=data_path, user=user, password=password)


def clean(self):
    """Placeholder routine that would normally return cleaned data

    Notes
    --------
    Supports 'clean', 'dusty', 'dirty' in the sense that it prints
    a message noting there is no cleaning.
    'None' is also supported as it signifies no cleaning.

    Routine is called by pysat, and not by the end user directly.

    """

    if self.clean_level in ['clean', 'dusty', 'dirty']:
        logger.warning('Generalized Madrigal data support has no cleaning.')

    return
